from __future__ import annotations

import logging
import os
import re
from dataclasses import dataclass
from typing import Any

import click
from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import SchemaDefinition, SlotDefinition
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView
from sqlalchemy import Column, ForeignKey, Index, MetaData, Numeric, Table, UniqueConstraint, create_mock_engine
from sqlalchemy.dialects.oracle import VARCHAR2
from sqlalchemy.types import Boolean, Date, DateTime, Enum, Float, Integer, Text, Time

from linkml._version import __version__
from linkml.transformers.relmodel_transformer import ForeignKeyPolicy, RelationalModelTransformer
from linkml.utils.generator import Generator, shared_arguments

logger = logging.getLogger(__name__)


class SqlNamingPolicy(Enum):
    preserve = "preserve"
    underscore = "underscore"
    camelcase = "camelcase"


# TODO: move this up
METAMODEL_TYPE_TO_BASE = {
    "string": "str",
    "integer": "int",
    "boolean": "Bool",
    "float": "float",
    "double": "double",
    "decimal": "Decimal",
    "time": "XSDTime",
    "date": "XSDDate",
    "datetime": "XSDDateTime",
    "uriorcurie": "URIorCURIE",
    "uri": "URI",
    "ncname": "NCName",
    "objectidentifier": "ElementIdentifier",
    "nodeidentifier": "NodeIdentifier",
}

RANGEMAP = {
    "str": Text(),
    "string": Text(),
    "String": Text(),
    "NCName": Text(),
    "URIorCURIE": Text(),
    "int": Integer(),
    "Decimal": Numeric(),
    "double": Float(),
    "float": Float(),
    "Bool": Boolean(),
    "URI": Text(),
    "XSDTime": Time(),
    "XSDDateTime": DateTime(),
    "XSDDate": Date(),
}


VARCHAR_REGEX = re.compile(r"VARCHAR2?(\((\d+)\))?")
ORACLE_MAX_VARCHAR_LENGTH = 4096


@dataclass
class SQLTableGenerator(Generator):
    """
    A :class:`~linkml.utils.generator.Generator` for creating SQL DDL.

    The basic algorithm for mapping a linkml schema S is as follows:

    - Each schema S corresponds to one database schema D (see SQLSchema)
    - Each Class C in S is mapped to a table T (see sqlalchemy.Table)
    - Each slot S in each C is mapped to a column Col (see sqlalchemy.Column)

    If the direct_mapping attribute is set to true, then no further transformations
    are applied. Note that this means:

    - inline objects are modeled as Text strings
    - multivalued fields are modeled as single Text strings

    This direct mapping is useful for simple spreadsheet/denormalized representations of complex data.
    However, for other applications, additional transformations should occur. These are:

    MULTIVALUED SLOTS

    The relational model does not have direct representation of lists. These are normalized as follows:

    If the range of the slot is a class, and there are no other slots whose range is this class,
    and the slot is for a class that has a singular primary key, then a backref is added.

    E.g. if we have User 0..* Address,
    then add a field User_id to Address.

    When SQLAlchemy bindings are created, a backref mapping is added.

    If the range of the slot is an enum or type, then a new linktable is created, and a backref added.

    E.g. if a class User has a multivalued slot alias whose range is a string,
    then create a table user_aliases, with two columns (1) alias [a string] and (2) a backref to user

    Each mapped slot C.S has a range R

    ranges that are types (literals):

    * If R is a type, and the slot is NOT multivalued, do a direct type conversion
    * If R is a type, and the slot is multivalued:

        * do not include the mapped column
        * create a new table T_S, with 2 columns: S, and a backref to T

    ranges that are classes:

    * Ref = map_class_to_table(R)
    * if R is a class, and the slot is NOT multivalued, and Ref has a singular primary key:

        * Col.type = ForeignKey(Ref.PK)

    * if R is a class, and the slot is NOT multivalued, and Ref has NO singular primary key:

        * add a foreign key C.pk to Ref
        * add a backref C.S => Ref, C.pk
        * remove Col from T

    * If R is a class, and the slot IS multivalued

    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["sql"]
    file_extension = "sql"
    uses_schemaloader = False

    # ObjectVars
    use_inherits: bool = False  # postgresql supports inheritance
    dialect: str = "sqlite"
    inject_primary_keys: bool = True
    use_foreign_keys: bool = True
    rename_foreign_keys: bool = False
    direct_mapping: bool = False
    relative_slot_num: bool = False
    default_length_oracle: int = ORACLE_MAX_VARCHAR_LENGTH
    generate_abstract_class_ddl: bool = True
    autogenerate_index: bool = True

    def serialize(self, **kwargs: dict[str, Any]) -> str:
        return self.generate_ddl(**kwargs)

    def generate_ddl(self, naming_policy: SqlNamingPolicy = None, **kwargs: dict[str, Any]) -> str:
        """
        Generate a DDL using the schema in self.schema.

        :param naming_policy: naming policy for columns, defaults to None
        :type naming_policy: SqlNamingPolicy, optional
        :return: the DDL as a string
        :rtype: str
        """
        ddl_str = ""

        def dump(sql, *multiparams, **params):
            nonlocal ddl_str
            ddl_str += f"{str(sql.compile(dialect=engine.dialect)).rstrip()};"

        engine = create_mock_engine(f"{self.dialect}://./MyDb", strategy="mock", executor=dump)
        schema_metadata = MetaData()
        sqltr = RelationalModelTransformer(SchemaView(self.schema))
        if not self.use_foreign_keys:
            sqltr.foreign_key_policy = ForeignKeyPolicy.NO_FOREIGN_KEYS
        tr_result = sqltr.transform(tgt_schema_name=kwargs.get("tgt_schema_name"), top_class=kwargs.get("top_class"))
        schema = tr_result.schema

        def sql_name(n: str) -> str:
            if not naming_policy or naming_policy == SqlNamingPolicy.preserve:
                return n
            if naming_policy == SqlNamingPolicy.underscore:
                return underscore(n)
            if naming_policy == SqlNamingPolicy.camelcase:
                return camelcase(n)
            msg = f"Unknown: {naming_policy}"
            raise Exception(msg)

        def strip_newlines(txt: str | None) -> str:
            if txt is None:
                return ""
            return txt.replace("\n", "")

        def strip_dangling_whitespace(txt: str) -> str:
            return "\n".join(line.rstrip() for line in txt.splitlines())

        # Currently SQLite dialect in SQLA does not generate comments; see
        # https://github.com/sqlalchemy/sqlalchemy/issues/1546#issuecomment-1067389172
        # As a workaround we add these as "--" comments via direct string manipulation
        include_comments = self.dialect == "sqlite"
        sv = SchemaView(schema)

        # This is a safeguard for checking duplicate item names within a table
        autogenerated_item_names = []

        # Iterate through the attributes in each class, creating Column objects.
        # This includes generating the appropriate column name, converting the range
        # into an SQL type, and adding a foreign key notation if appropriate.
        for cn, c in schema.classes.items():
            desc = strip_newlines(c.description)
            if include_comments:
                if c.abstract:
                    if desc:
                        ddl_str += f"-- # Abstract Class: {cn} Description: {desc}\n"
                    else:
                        ddl_str += f"-- # Abstract Class: {cn}\n"
                else:
                    if desc:
                        ddl_str += f"-- # Class: {cn} Description: {desc}\n"
                    else:
                        ddl_str += f"-- # Class: {cn}\n"
            pk_slot = sv.get_identifier_slot(cn)
            if c.attributes:
                cols = []
                for sn, s in c.attributes.items():
                    is_pk = "primary_key" in s.annotations
                    if pk_slot:
                        is_pk = sn == pk_slot.name
                    # else:
                    #    is_pk = True  ## TODO: use unique key
                    args = []
                    if s.range in schema.classes and self.use_foreign_keys:
                        fk = sql_name(self.get_id_or_key(s.range, sv))
                        args = [ForeignKey(fk)]
                    field_type = self.get_sql_range(s, schema)
                    fk_index_cond = (s.key or s.identifier) and self.autogenerate_index
                    pk_index_cond = is_pk and self.autogenerate_index
                    is_index = fk_index_cond or pk_index_cond
                    col = Column(
                        sql_name(sn),
                        field_type,
                        *args,
                        primary_key=is_pk,
                        nullable=not s.required,
                        index=is_index,
                    )
                    if is_index:
                        # this is what SQLAlchemy generates, TODO: perhaps we need to expand this
                        autogenerated_item_names.append(sql_name("ix_" + sql_name(cn) + "_" + sql_name(sn)))
                    if include_comments:
                        desc = strip_newlines(s.description)
                        comment_str = f"--     * Slot: {sn}"
                        if desc:
                            comment_str += f" Description: {desc}"
                        ddl_str += comment_str + "\n"
                    if s.description:
                        col.comment = s.description
                    cols.append(col)
                # convert unique_keys into a uniqueness constraint for the table
                for uc in c.unique_keys.values():
                    sql_names = [sql_name(sn) if sn in c.attributes else None for sn in uc.unique_key_slots]
                    if any(sn is None for sn in sql_names):
                        continue
                    sql_uc = UniqueConstraint(*sql_names)
                    cols.append(sql_uc)
                    # Anything that has a unique constraint should have an associated index with it
                    if self.autogenerate_index:
                        uc_index_name = sql_name(cn)
                        for name in sql_names:
                            uc_index_name = uc_index_name + "_" + name
                        uc_index_name = uc_index_name + "_idx"
                        is_duplicate = self.check_duplicate_entry_names(
                            autogenerated_item_names, sql_name(uc_index_name)
                        )
                        if not is_duplicate:
                            sql_names = [sql_name(uc_index_name)] + sql_names
                            uc_index = Index(*sql_names)
                            cols.append(uc_index)
                            autogenerated_item_names.append(sql_name(uc_index_name))
            if not c.abstract or (c.abstract and self.generate_abstract_class_ddl):
                for tag, annotation in sorted(c.annotations.items()):
                    if tag == "index":
                        value_dict = {k: annotation for k, annotation in annotation.value._items()}
                        for key, value in value_dict.items():
                            test_name = sql_name(key)
                            name_exists = self.check_duplicate_entry_names(autogenerated_item_names, test_name)
                            if not name_exists:
                                value_sql_name = [sql_name(col_name) for col_name in value]
                                annotation_ind = Index(test_name, *value_sql_name)
                                autogenerated_item_names.append(test_name)
                                cols.append(annotation_ind)
                Table(sql_name(cn), schema_metadata, *cols, comment=str(c.description))
        schema_metadata.create_all(engine)
        ddl_str = strip_dangling_whitespace(ddl_str)
        if ddl_str[-1:] != "\n":
            ddl_str += "\n"
        return ddl_str

    def get_oracle_sql_range(self, slot: SlotDefinition) -> Text | VARCHAR2 | None:
        """
        Generate the appropriate range for Oracle SQL.

        :param slot: the slot under examination
        :type slot: SlotDefinition
        :return: appropriate type for Oracle SQL or None
        :rtype: Text | VARCHAR2 | None
        """
        slot_range = slot.range
        if not slot_range:
            return None

        if slot_range.lower() in ["str", "string"]:
            # string type data should be represented as a VARCHAR2
            return VARCHAR2(self.default_length_oracle)

        # check whether the slot range matches the regex "VARCHAR2?(\((\d+)\))?"
        match = re.match(VARCHAR_REGEX, slot_range)
        if match:
            # match.group(2) is the digits in brackets after VARCHAR
            # i.e. a defined length for the VARCHAR
            if match.group(2) and int(match.group(2)) > ORACLE_MAX_VARCHAR_LENGTH:
                msg = (
                    "Warning: range exceeds maximum Oracle VARCHAR length, "
                    f"CLOB type will be returned: {slot_range} for {slot.name} = {slot.range}"
                )
                logger.info(msg)
                return Text()
            # set the length to either the varchar length (as defined in the slot_range)
            # or the default
            return VARCHAR2(match.group(2) or self.default_length_oracle)

        # use standard SQL range matching for anything else
        return None

    def get_sql_range(self, slot: SlotDefinition, schema: SchemaDefinition = None):
        """Get the slot range as a SQL Alchemy column type."""
        slot_range = slot.range

        if self.dialect == "oracle":
            range_type = self.get_oracle_sql_range(slot)
            if range_type:
                return range_type

        if slot_range is None:
            return Text()

        # if no SchemaDefinition is explicitly provided as an argument
        # then simply use the schema that is provided to the SQLTableGenerator() object
        if not schema:
            schema = self.schema

        sv = SchemaView(schema)
        if slot_range in sv.all_classes():
            # FK type should be the same as the identifier of the foreign key
            fk = sv.get_identifier_slot(slot_range)
            if fk:
                return self.get_sql_range(fk, sv.schema)
            return Text()

        if slot_range in sv.all_enums():
            e = sv.all_enums()[slot_range]
            if e.permissible_values is not None:
                vs = [str(v) for v in e.permissible_values]
                return Enum(name=e.name, *vs)

        if slot_range in METAMODEL_TYPE_TO_BASE:
            range_base = METAMODEL_TYPE_TO_BASE[slot_range]
        elif slot_range in sv.all_types():
            range_base = sv.all_types()[slot_range].base
        else:
            logger.error(f"Unknown range: {slot_range} for {slot.name} = {slot.range}")
            return Text()

        if range_base in RANGEMAP:
            return RANGEMAP[range_base]

        logger.error(f"Unknown range base: {range_base} for {slot.name} = {slot.range}")
        return Text()

    @staticmethod
    def get_id_or_key(cn: str, sv: SchemaView) -> str:
        """Given a named class, retrieve the identifier or key slot."""
        pk = sv.get_identifier_slot(cn, use_key=True)
        if pk is None:
            msg = f"No PK for {cn}"
            raise Exception(msg)
        pk_name = pk.alias if pk.alias else pk.name
        return f"{cn}.{pk_name}"

    # Scans the schema for duplicate names that exist as a best practice
    @staticmethod
    def check_duplicate_entry_names(autogen: list, item_name: str) -> bool:
        if item_name in autogen:
            msg = f"Warning: {item_name} already exists, please generate a new name"
            logger.warning(msg)
            return True
        return False


@shared_arguments(SQLTableGenerator)
@click.command(name="sqltables")
@click.option(
    "--dialect",
    default="sqlite",
    show_default=True,
    help="SQL-Alchemy dialect, e.g. sqlite, mysql+odbc",
)
@click.option("--sqla-file", help="Path to sqlalchemy generated python")
@click.option(
    "--relmodel-output",
    help="Path to intermediate LinkML YAML of transformed relational model",
)
@click.option("--python-import", help="Python import header for generated sql-alchemy code")
@click.option(
    "--use-foreign-keys/--no-use-foreign-keys",
    default=True,
    show_default=True,
    help="Emit FK declarations",
)
@click.option(
    "--default_length_oracle",
    default=ORACLE_MAX_VARCHAR_LENGTH,
    show_default=True,
    help="Default length of varchar based arguments for oracle dialects",
)
@click.option(
    "--autogenerate_index",
    default=True,
    show_default=True,
    help="Enable the creation of indexes on all columns generated",
)
@click.option(
    "--generate_abstract_class_ddl",
    default=True,
    show_default=True,
    help="A manual override to omit the abstract classes, set to true as a default for testing sake",
)
@click.version_option(__version__, "-V", "--version")
def cli(
    yamlfile: str,
    relmodel_output: str,
    sqla_file: str | None = None,
    python_import: str = None,
    dialect: str | None = None,
    use_foreign_keys: bool = True,
    **args,
):
    """Generate SQL DDL representation."""
    if relmodel_output:
        sv = SchemaView(yamlfile)
        rtr = RelationalModelTransformer(sv)
        if not use_foreign_keys:
            rtr.foreign_key_policy = ForeignKeyPolicy.NO_FOREIGN_KEYS
        rtr_result = rtr.transform("foo")
        relmodel_schema = rtr_result.schema
        yaml_dumper.dump(relmodel_schema, to_file=relmodel_output)
    gen = SQLTableGenerator(yamlfile, use_foreign_keys=use_foreign_keys, **args)
    if dialect:
        gen.dialect = dialect
    print(gen.generate_ddl())
    if sqla_file is not None:
        raise NotImplementedError("SQLAGen not implemented")


if __name__ == "__main__":
    cli()



import os
from collections import defaultdict
from typing import Optional

from linkml_runtime.dumpers import CSVDumper, JSONDumper, RDFLibDumper, TSVDumper, YAMLDumper
from linkml_runtime.linkml_model.meta import ClassDefinitionName, SlotDefinitionName
from linkml_runtime.loaders import CSVLoader, JSONLoader, RDFLibLoader, TSVLoader, YAMLLoader
from linkml_runtime.loaders.loader_root import Loader
from linkml_runtime.utils.schemaview import SchemaView

from linkml.generators.jsonldcontextgen import ContextGenerator

dumpers_loaders = {
    "yml": (YAMLDumper, YAMLLoader),
    "yaml": (YAMLDumper, YAMLLoader),
    "json": (JSONDumper, JSONLoader),
    "rdf": (RDFLibDumper, RDFLibLoader),
    "ttl": (RDFLibDumper, RDFLibLoader),
    "json-ld": (JSONDumper, JSONLoader),
    "csv": (CSVDumper, CSVLoader),
    "tsv": (TSVDumper, TSVLoader),
}

aliases = {
    "rdf": "ttl",
    "jsonld": "json-ld",
}


def _get_format(path: str, specified_format: str = None, default=None):
    if specified_format is None:
        if path is None:
            if default is None:
                raise Exception("Must pass format option OR pass a filename with known file suffix")
            else:
                specified_format = default
        else:
            _, ext = os.path.splitext(path)
            if ext is not None:
                specified_format = ext.replace(".", "")
            else:
                raise Exception(f"Must pass format option OR use known file suffix: {path}")
    specified_format = specified_format.lower()
    if specified_format in aliases:
        specified_format = aliases[specified_format]
    return specified_format


def _is_xsv(fmt: str) -> bool:
    return fmt == "csv" or fmt == "tsv"


def _is_rdf_format(fmt: str) -> bool:
    return fmt == "rdf" or fmt == "ttl" or fmt == "turtle" or fmt == "json-ld"


def get_loader(fmt: str) -> Loader:
    return dumpers_loaders[fmt][1]()


def get_dumper(fmt: str) -> Loader:
    return dumpers_loaders[fmt][0]()


def _get_context(schema) -> str:
    return ContextGenerator(schema).serialize()


def infer_root_class(sv: SchemaView) -> Optional[ClassDefinitionName]:
    """
    Infer the class that should be at the root of the object tree

    (Note this is distinct from the root of the class hierarchy)

    If a class is explicitly designated with tree_root, use this.
    Otherwise use the class that is not referenced as a range in any other class.
    """
    for c in sv.all_classes().values():
        if c.tree_root:
            return c.name
    refs = defaultdict(int)
    for cn in sv.all_classes().keys():
        for sn in sv.class_slots(cn):
            slot = sv.induced_slot(sn, cn)
            r = slot.range
            if r in sv.all_classes():
                for a in sv.class_ancestors(r):
                    refs[a] += 1
    candidates = [cn for cn in sv.all_classes().keys() if cn not in refs]

    # throw Exception if unambiguous root cannot be inferred
    if len(candidates) > 1:
        raise RuntimeError(
            f"Multiple potential target classes found: {candidates}. "
            "Please specify a target using --target-class or by adding "
            "tree_root: true to the relevant class in the schema."
        )

    if len(candidates) == 1:
        return candidates[0]
    else:
        return None


def infer_index_slot(sv: SchemaView, root_class: ClassDefinitionName) -> Optional[SlotDefinitionName]:
    index_slots = []
    for sn in sv.class_slots(root_class):
        slot = sv.induced_slot(sn, root_class)
        if slot.multivalued and slot.range in sv.all_classes():
            index_slots.append(sn)
    if len(index_slots) == 1:
        return index_slots[0]
    else:
        return None



import abc
import re
import sys
from abc import ABC
from typing import Any, Optional, Union

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import (
    Boolean,
    ClassDefinition,
    Date,
    Datetime,
    Decimal,
    Double,
    EnumDefinitionName,
    Float,
    Integer,
    SlotDefinition,
    String,
    Time,
    Uri,
    types,
)
from linkml_runtime.linkml_model.types import (
    Curie,
    DateOrDatetime,
    Jsonpath,
    Jsonpointer,
    Ncname,
    Nodeidentifier,
    Objectidentifier,
    Sparqlpath,
    Uriorcurie,
)

TYPES_TYPE: TypeAlias = Union[
    type[Boolean],
    type[Curie],
    type[Date],
    type[DateOrDatetime],
    type[Datetime],
    type[Decimal],
    type[Double],
    type[Float],
    type[Integer],
    type[Jsonpath],
    type[Jsonpointer],
    type[Ncname],
    type[Nodeidentifier],
    type[Objectidentifier],
    type[Sparqlpath],
    type[String],
    type[Time],
    type[Uri],
    type[Uriorcurie],
]

TYPES = [
    t
    for t in types.__dict__.values()
    if isinstance(t, type) and t.__module__ == types.__name__ and hasattr(t, "type_name")
]


class IfAbsentProcessor(ABC):
    """
    Processes value of ifabsent slot.

    See `<https://w3id.org/linkml/ifabsent>`_.
    """

    ifabsent_regex = re.compile(r"""(?:(?P<type>\w+)\()?[\"\']?(?P<default_value>[^\(\)\"\')]*)[\"\']?\)?""")

    def __init__(self, schema_view: SchemaView):
        self.schema_view = schema_view

    def process_slot(self, slot: SlotDefinition, cls: ClassDefinition) -> Optional[str]:
        if slot.ifabsent:
            ifabsent_match = self.ifabsent_regex.search(slot.ifabsent)
            ifabsent_default_value = ifabsent_match.group("default_value")

            return self._map_to_default_value(slot, ifabsent_default_value, cls)

        return None

    def _map_to_default_value(
        self, slot: SlotDefinition, ifabsent_default_value: Any, cls: ClassDefinition
    ) -> Optional[str]:
        # Used to manage specific cases that aren't generic
        mapped, custom_default_value = self.map_custom_default_values(ifabsent_default_value, slot, cls)
        if mapped:
            return custom_default_value

        base_type = self._base_type(slot.range)

        if base_type is String:
            return self.map_string_default_value(ifabsent_default_value, slot, cls)

        if base_type is Boolean:
            if re.match(r"^[Tt]rue$", ifabsent_default_value):
                return self.map_boolean_true_default_value(slot, cls)
            elif re.match(r"^[Ff]alse$", ifabsent_default_value):
                return self.map_boolean_false_default_value(slot, cls)
            else:
                raise ValueError(
                    f"The ifabsent value `{slot.ifabsent}` of the `{slot.name}` slot does not match a valid boolean "
                    f"value"
                )

        if base_type is Integer:
            return self.map_integer_default_value(ifabsent_default_value, slot, cls)

        if base_type is Float:
            return self.map_float_default_value(ifabsent_default_value, slot, cls)

        if base_type is Double:
            return self.map_double_default_value(ifabsent_default_value, slot, cls)

        if base_type is Decimal:
            return self.map_decimal_default_value(ifabsent_default_value, slot, cls)

        if base_type is Time:
            match = re.match(r"^(\d{2}):(\d{2}):(\d{2}).*$", ifabsent_default_value)
            if match:
                return self.map_time_default_value(match[1], match[2], match[3], slot, cls)
            else:
                raise ValueError(
                    f"The ifabsent value `{slot.ifabsent}` of the `{slot.name}` slot does not match a valid time value"
                )

        # TODO manage timezones and offsets
        if base_type is Date:
            match = re.match(r"^(\d{4})-(\d{2})-(\d{2})$", ifabsent_default_value)
            if match:
                return self.map_date_default_value(match[1], match[2], match[3], slot, cls)
            else:
                raise ValueError(
                    f"The ifabsent value `{slot.ifabsent}` of the `{slot.name}` slot does not match a valid date value"
                )

        # TODO manage timezones and offsets
        if base_type is Datetime:
            match = re.match(r"^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}).*$", ifabsent_default_value)
            if match:
                return self.map_datetime_default_value(
                    match[1], match[2], match[3], match[4], match[5], match[6], slot, cls
                )
            else:
                raise ValueError(
                    f"The ifabsent value `{slot.ifabsent}` of the `{slot.name}` slot does not match a valid datetime "
                    f"value"
                )

        # TODO manage timezones and offsets
        if base_type is DateOrDatetime:
            match = re.match(r"^(\d{4})-(\d{2})-(\d{2})(?:T(\d{2}):(\d{2}):(\d{2}))?.*$", ifabsent_default_value)
            if match and (match[4] is None or match[5] is None or match[6] is None):
                return self.map_date_default_value(match[1], match[2], match[3], slot, cls)
            elif match:
                return self.map_datetime_default_value(
                    match[1], match[2], match[3], match[4], match[5], match[6], slot, cls
                )
            else:
                raise ValueError(
                    f"The ifabsent value `{slot.ifabsent}` of the `{slot.name}` slot does not match a valid date or "
                    f"datetime value"
                )

        if base_type is Uri:
            return self.map_uri_default_value(ifabsent_default_value, slot, cls)

        if base_type is Curie:
            return self.map_curie_default_value(ifabsent_default_value, slot, cls)

        if base_type is Uriorcurie:
            return self.map_uri_or_curie_default_value(ifabsent_default_value, slot, cls)

        if base_type is Ncname:
            return self.map_nc_name_default_value(ifabsent_default_value, slot, cls)

        if base_type is Objectidentifier:
            return self.map_object_identifier_default_value(ifabsent_default_value, slot, cls)

        if base_type is Nodeidentifier:
            return self.map_node_identifier_default_value(ifabsent_default_value, slot, cls)

        if base_type is Jsonpointer:
            return self.map_json_pointer_default_value(ifabsent_default_value, slot, cls)

        if base_type is Jsonpath:
            return self.map_json_path_default_value(ifabsent_default_value, slot, cls)

        if base_type is Sparqlpath:
            return self.map_sparql_path_default_value(ifabsent_default_value, slot, cls)

        # -----------------------
        #    Enum slot ranges
        # -----------------------

        for enum_name, enum in self.schema_view.all_enums().items():
            if enum_name == slot.range:
                for permissible_value_name, permissible_value in enum.permissible_values.items():
                    if permissible_value_name == ifabsent_default_value:
                        return self.map_enum_default_value(enum_name, permissible_value_name, slot, cls)

        raise ValueError(f"The ifabsent value `{slot.ifabsent}` of the `{slot.name}` slot could not be processed")

    def _base_type(self, range_: str) -> Optional[TYPES_TYPE]:
        """
        Find the linkml base type that corresponds to either a matching type_name or custom type

        First check for an explicit match of the range == TypeDefinition.type_name
        Then check for explicit inheritance via typeof
        Finally check for implicit matching via matching base

        Don't raise here, just return None in case another method of resolution like enums are
        available
        """
        # first check for matching type using type_name - ie. range is already a base type

        for typ in TYPES:
            if range_ == typ.type_name:
                return typ

        # if we're not a type, return None to indicate that, e.g. if an enum's permissible_value
        if range_ not in self.schema_view.all_types(imports=True):
            return

        # then check explicit descendents of types
        # base types do not inherit from one another, so the last ancestor is always a base type
        # if it is inheriting from a base type
        ancestor = self.schema_view.type_ancestors(range_)[-1]
        for typ in TYPES:
            if ancestor == typ.type_name:
                return typ

        # finally check if we have a matching base
        induced_typ = self.schema_view.induced_type(range_)
        if induced_typ.repr is None and induced_typ.base is None:
            return None
        for typ in TYPES:
            # types always inherit from a single type, and that type is their base
            # our range can match it with repr or base
            typ_base = typ.__mro__[1].__name__
            if typ_base == induced_typ.base:
                return typ

    @abc.abstractmethod
    def map_custom_default_values(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition) -> (bool, str):
        """
        Maps custom default values that aren't generic behaviours.

        @param default_value: the default value extracted from the ifabsent attribute
        @param slot: the definition of the slot
        @param cls: the definition of the class
        @return: a boolean that indicates if the value has been mapped followed by the mapped value
        """
        return False, None

    @abc.abstractmethod
    def map_string_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_integer_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_boolean_true_default_value(self, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_boolean_false_default_value(self, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_float_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_double_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_decimal_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_time_default_value(self, hour: str, minutes: str, seconds: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_date_default_value(self, year: str, month: str, day: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_datetime_default_value(
        self,
        year: str,
        month: str,
        day: str,
        hour: str,
        minutes: str,
        seconds: str,
        slot: SlotDefinition,
        cls: ClassDefinition,
    ):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_uri_or_curie_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_curie_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_uri_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_nc_name_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_object_identifier_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_node_identifier_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_json_pointer_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_json_path_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_sparql_path_default_value(self, default_value: str, slot: SlotDefinition, cls: ClassDefinition):
        raise NotImplementedError()

    @abc.abstractmethod
    def map_enum_default_value(
        self, enum_name: EnumDefinitionName, permissible_value_name: str, slot: SlotDefinition, cls: ClassDefinition
    ):
        raise NotImplementedError()

    def _uri_for(self, s: str) -> str:
        uri = str(self.schema_view.namespaces().uri_for(s))
        return self.schema_view.namespaces().curie_for(uri, True, True) or self._strval(uri)

    def _strval(self, txt: str) -> str:
        txt = str(txt).replace('"', '\\"')
        return f'"{txt}"'



"""
flatten_transformer.py

This module provides a FlattenTransformer that transforms a LinkML schema into a flattened version.
The transformer takes a target class name and rolls up all slots from its descendants into that class.
This is useful for implementing single table inheritance or simplifying inheritance for systems like KGX.
"""

import copy
from dataclasses import dataclass
from typing import Optional

from linkml_runtime.linkml_model.meta import (
    ClassDefinitionName,
    SchemaDefinition,
)
from linkml_runtime.utils.schemaview import SchemaView

from linkml.transformers.model_transformer import ModelTransformer


@dataclass
class FlattenTransformerConfiguration:
    """
    Configuration options for the FlattenTransformer.

    Attributes:
        preserve_class_designator: If True, preserves the original class name in the specified
            class designator slot (e.g., 'category').
        class_designator_slot: The name of the slot that designates the original class (e.g., 'category').
        include_all_classes: If True, keeps all classes in the schema; otherwise, only the target class
            and non-descendants are kept.
        include_mixins: If True, includes properties from mixins in the flattened class.
    """

    preserve_class_designator: bool = True
    class_designator_slot: Optional[str] = "category"
    include_all_classes: bool = False
    include_mixins: bool = True


class RollupTransformer(ModelTransformer):
    """
    Transforms a LinkML schema by flattening an inheritance hierarchy into a single class.

    This transformer takes a schema and a target class name, and produces a new schema where
    all descendant classes of the target class are removed and their slots are rolled up into the
    target class.
    """

    def __init__(
        self,
        target_class: str,
        config: Optional[FlattenTransformerConfiguration] = None,
    ):
        super().__init__()
        self.target_class = target_class
        self.config = config or FlattenTransformerConfiguration()
        self._descendant_classes: set[str] = set()
        self._class_to_slot_map: dict[str, set[str]] = {}

    def transform(self, tgt_schema_name: str = None, top_class: ClassDefinitionName = None) -> SchemaDefinition:
        """
        Transform the schema by flattening the inheritance hierarchy.

        Args:
            tgt_schema_name: Optional target schema name (not used in this transformer)
            top_class: Optional top class name (not used in this transformer)

        Returns:
            A transformed SchemaDefinition with the target class having all slots from its descendants.
        """
        if not self.schemaview:
            raise ValueError("SchemaView not set. Ensure schema is set by calling set_schema().")

        schema_copy = copy.deepcopy(self.source_schema)
        view = SchemaView(schema_copy)
        descendants = view.class_descendants(self.target_class, imports=True, mixins=True, reflexive=False, is_a=True)
        self._descendant_classes = set(descendants)

        # Collect all slots from the target class and its descendant classes, and
        # populate a mapping of class name to the set of its slot names.
        for descendant in self._descendant_classes:
            slots = view.class_induced_slots(descendant)
            self._class_to_slot_map[descendant] = {s.name for s in slots}

        # flatten the schema
        return self._create_flattened_schema(schema_copy)

    def _create_flattened_schema(self, schema: SchemaDefinition) -> SchemaDefinition:
        """
        Create a new flattened version of the schema based on collected slots.
        """
        target_class_def = schema.classes[self.target_class]
        existing_slot_names = set(target_class_def.slots or [])
        all_slots = set()

        for descendant in self._descendant_classes:
            if descendant in schema.classes:
                for slot_name in self._class_to_slot_map.get(descendant, set()):
                    if slot_name not in existing_slot_names:
                        all_slots.add(slot_name)

        if not target_class_def.slots:
            target_class_def.slots = []

        target_class_def.slots.extend(list(all_slots))

        # Optionally remove descendant classes.
        if not self.config.include_all_classes:
            for descendant in self._descendant_classes:
                if descendant in schema.classes:
                    del schema.classes[descendant]

        return schema



from typing import IO, Any, Optional

import click

from ..linter import LinterProblem


class Formatter:
    def __init__(self, file: Optional[IO[Any]] = None) -> None:
        self.file = file

    def write(self, message: str):
        click.echo(message, file=self.file)

    def start_report(self):
        pass

    def start_schema(self, name: str):
        pass

    def handle_problem(self, problem: LinterProblem):
        pass

    def end_schema(self):
        pass

    def end_report(self):
        pass



import logging
import os
from collections import OrderedDict
from collections.abc import Iterator, Mapping
from copy import deepcopy
from pathlib import Path
from typing import Optional, TextIO, Union, cast
from urllib.parse import urlparse

from jsonasobj2 import values
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    ElementName,
    EnumDefinition,
    EnumDefinitionName,
    SchemaDefinition,
    SlotDefinition,
    SlotDefinitionName,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.context_utils import parse_import_map
from linkml_runtime.utils.formatutils import camelcase, mangled_attribute_name, sfx, underscore
from linkml_runtime.utils.metamodelcore import Bool
from linkml_runtime.utils.namespaces import Namespaces
from linkml_runtime.utils.yamlutils import TypedNode

from linkml.utils.mergeutils import merge_classes, merge_schemas, merge_slots, slot_usage_name
from linkml.utils.rawloader import load_raw_schema
from linkml.utils.schemasynopsis import SchemaSynopsis

lgr = logging.getLogger(__name__)


class SchemaLoader:
    def __init__(
        self,
        data: Union[str, TextIO, SchemaDefinition, dict, Path],
        base_dir: Optional[str] = None,
        namespaces: Optional[Namespaces] = None,
        useuris: Optional[bool] = None,
        importmap: Optional[Mapping[str, str]] = None,
        logger: Optional[logging.Logger] = None,
        mergeimports: Optional[bool] = True,
        emit_metadata: Optional[bool] = True,
        source_file_date: Optional[str] = None,
        source_file_size: Optional[int] = None,
    ) -> None:
        """Constructor - load and process a YAML or pre-processed schema

        :param data: YAML schema text, python dict loaded from yaml,  URL, file name, open file or SchemaDefinition
        :param base_dir: base directory or URL where Schema came from
        :param namespaces: namespaces collector
        :param useuris: True means class_uri and slot_uri are identifiers.  False means they are mappings.
        :param importmap: A map from import entries to URI or file name.
        :param logger: Target Logger, if any
        :param mergeimports: True means combine imports into single package. False means separate packages
        :param emit_metadata: True means include source file, size and date
        :param source_file_date: modification of source file
        :param source_file_size: size of source file
        """
        self.logger = logger if logger is not None else lgr
        if isinstance(data, SchemaDefinition):
            self.schema = data
        else:
            self.schema = load_raw_schema(
                data,
                base_dir=base_dir,
                merge_modules=mergeimports,
                source_file_date=source_file_date,
                source_file_size=source_file_size,
            )
        # Map from URI to source and version tuple
        self.loaded: OrderedDict[str, tuple[str, str]] = {
            self.schema.id: (self.schema.source_file, self.schema.version)
        }
        self.base_dir = self._get_base_dir(base_dir)
        self.namespaces = namespaces if namespaces else Namespaces()
        self.useuris = useuris if useuris is not None else True
        self.importmap = parse_import_map(importmap, self.base_dir) if importmap is not None else dict()
        self.source_file_date = source_file_date
        self.source_file_size = source_file_size
        self.synopsis: Optional[SchemaSynopsis] = None
        self.schema_location: Optional[str] = None
        self.schema_defaults: dict[str, str] = {}  # Map from schema URI to default namespace
        self.merge_modules = mergeimports
        self.emit_metadata = emit_metadata

    def resolve(self) -> SchemaDefinition:
        """Reconcile a loaded schema, applying is_a, mixins, apply_to's and other such things.  Also validate the
        content and load a SchemaSynopsis entry

        :return: Fully resolved definition
        """
        if not self.schema.default_range:
            self.schema.default_range = "string"
            self.logger.info(f"Default_range not specified. Default set to '{self.schema.default_range}'")

        # Process the namespace declarations
        if not self.schema.default_prefix:
            self.schema.default_prefix = sfx(self.schema.id)
        self.schema_defaults[self.schema.id] = self.schema.default_prefix
        for prefix in self.schema.prefixes.values():
            self.namespaces[prefix.prefix_prefix] = prefix.prefix_reference
        for cmap in self.schema.default_curi_maps:
            self.namespaces.add_prefixmap(cmap, include_defaults=False)

        # Process imports
        for imp in self.schema.imports:
            sname = self.importmap.get(str(imp), imp)  # Import map may use CURIE
            # substitute CURIE only if we don't have a local file name with drive letter (windows)
            if not os.path.splitdrive(sname)[0]:
                if ":" in sname:
                    # allow mapping of a prefix to a folder/directory
                    toks = sname.split(":")
                    pfx = toks[0]
                    if pfx in self.importmap:
                        sname = os.path.join(self.importmap[pfx], ":".join(toks[1:]))
                    else:
                        sname = self.namespaces.uri_for(sname)
            sname = self.importmap.get(str(sname), sname)  # It may also use URI or other forms
            import_schemadefinition = load_raw_schema(
                sname + ".yaml",
                base_dir=os.path.dirname(self.schema.source_file) if self.schema.source_file else self.base_dir,
                merge_modules=self.merge_modules,
                emit_metadata=self.emit_metadata,
            )
            loaded_schema = (str(sname), import_schemadefinition.version)
            if import_schemadefinition.id in self.loaded:
                # If we've already loaded this, make sure that we've got the same version
                if self.loaded[import_schemadefinition.id][1] != loaded_schema[1]:
                    self.raise_value_error(
                        f"Schema {import_schemadefinition.name} - version mismatch",
                        import_schemadefinition.name,
                    )
                # Note: for debugging purposes we also check whether the version
                #       came from the same spot.  This should be loosened to
                #       version only once we're sure that everything is working
                # TODO: The test below needs review -- there are cases where it
                #       fails because self.loaded[...][0] has the full path name
                #       and loaded_schema[0] is just the local name
                # if self.loaded[import_schemadefinition.id] != loaded_schema:
                #     self.raise_value_error(f"Schema imported from different files: "
                #                            f"{self.loaded[import_schemadefinition.id][0]} : {loaded_schema[0]}")
            else:
                self.loaded[import_schemadefinition.id] = loaded_schema
                merge_schemas(
                    self.schema,
                    import_schemadefinition,
                    imp,
                    self.namespaces,
                    merge_imports=self.merge_modules,
                )
                self.schema_defaults[import_schemadefinition.id] = import_schemadefinition.default_prefix

        if not self.namespaces._default:
            if "://" in self.schema.default_prefix:
                self.namespaces._default = self.schema.default_prefix
            elif self.schema.default_prefix in self.namespaces:
                self.namespaces._default = self.namespaces[self.schema.default_prefix]
            else:
                self.raise_value_error(
                    f"Default prefix: {self.schema.default_prefix} is not defined",
                    self.schema.default_prefix,
                )

        self.namespaces._base = (
            self.schema.default_prefix
            if ":" in self.schema.default_prefix
            else self.namespaces[self.schema.default_prefix]
        )

        # Promote embedded attribute definitions to first class slots.
        for cls in self.schema.classes.values():
            for attribute in cls.attributes.values():
                mangled_slot_name = mangled_attribute_name(cls.name, attribute.name)
                if mangled_slot_name in self.schema.slots:
                    # mangled names are overwritten if a schema with attributes is passed in
                    # TODO: handle this in a more graceful way
                    #  see https://github.com/linkml/linkml/issues/872
                    self.logger.warning(
                        f'Class: "{cls.name}" attribute "{attribute.name}" - '
                        f"mangled name: {mangled_slot_name} already exists",
                    )
                new_slot = SlotDefinition(**attribute.__dict__)
                new_slot.domain_of.append(cls.name)
                new_slot.imported_from = cls.imported_from
                new_slot.from_schema = cls.from_schema
                if not new_slot.alias:
                    new_slot.alias = attribute.name
                new_slot.name = mangled_slot_name
                self.schema.slots[new_slot.name] = new_slot
                cls.slots.append(mangled_slot_name)

        # Assign class slot ownership
        for cls in self.schema.classes.values():
            if not isinstance(cls, ClassDefinition):
                name = cls["name"] if "name" in cls else "Unknown"
                self.raise_value_error(
                    f'Class "{name} (type: {type(cls)})" definition is not a class definition',
                    name,
                )
            if isinstance(cls.slots, str):
                self.logger.warning(f"File: {self.schema.source_file} Class: {cls.name} Slots are not an array")
                cls.slots = [cls.slots]
            for slotname in cls.slots:
                if slotname in self.schema.slots:
                    slot = self.schema.slots[cast(SlotDefinitionName, slotname)]
                    slot.owner = cls.name
                    if cls.name not in slot.domain_of:
                        slot.domain_of.append(cls.name)
                else:
                    self.raise_value_error(f'Class "{cls.name}" - unknown slot: "{slotname}"', slotname)

        # Process slots defined as slot usages
        self.process_slot_usage_definitions()

        # Massage initial set of slots
        for slot in self.schema.slots.values():
            # Propagate domain to containing class
            if slot.domain and slot.domain in self.schema.classes:
                if slot.name not in self.schema.classes[slot.domain].slots:
                    slot.owner = slot.name
                    # self.schema.classes[slot.domain].slots.append(slot.name)
            elif slot.domain:
                self.raise_value_error(
                    f"slot: {slot.name} - unrecognized domain ({slot.domain})",
                    slot.domain,
                )

            # Validate the slot range
            if (
                slot.range is not None
                and slot.range not in self.schema.types
                and slot.range not in self.schema.classes
                and slot.range not in self.schema.enums
            ):
                self.raise_value_error(f"slot: {slot.name} - unrecognized range ({slot.range})", slot.range)

            # check constraints for usage of equals_string and equals_string_in
            self._check_equals_string(slot)

        # apply to --> mixins
        for cls in self.schema.classes.values():
            for apply_to_cls in cls.apply_to:
                if apply_to_cls in self.schema.classes:
                    self.schema.classes[apply_to_cls].mixins.append(cls.name)
                else:
                    self.raise_value_error(
                        f'Class "{cls.name}" unknown apply_to target: {apply_to_cls}',
                        apply_to_cls,
                    )
            # Class URI's also count as (trivial) mappings
            if cls.class_uri is not None:
                cls.mappings.insert(0, cls.class_uri)
            if cls.class_uri is None or not self.useuris:
                from_schema = cls.from_schema
                if from_schema is None:
                    from_schema = self.schema.id
                # if cls.from_schema is None:
                #    raise Exception(f"Class has no from_schema: {cls}")
                suffixed_cls_schema = sfx(from_schema)
                cls.class_uri = self.namespaces.uri_or_curie_for(
                    self.schema_defaults.get(cls.from_schema, suffixed_cls_schema),
                    camelcase(cls.name),
                )

        # Get the inverse ducks all in a row before we start filling other stuff in
        for slot in self.schema.slots.values():
            if slot.inverse:
                inverse_slot = self.schema.slots.get(slot.inverse, None)
                if inverse_slot:
                    if not inverse_slot.inverse:
                        inverse_slot.inverse = slot.name
                    elif inverse_slot.inverse != slot.name:
                        self.raise_value_error(
                            f"Slot {slot.name}.inverse ({slot.inverse}) does not match "
                            f"slot {inverse_slot.name}.inverse ({inverse_slot.inverse})"
                        )
                else:
                    self.raise_value_error(f"Slot {slot.name}.inverse ({slot.inverse}) is not defined")

        # Update slots with parental information
        merged_slots: list[SlotDefinitionName] = []
        for slot in self.schema.slots.values():
            if not slot.from_schema:
                slot.from_schema = self.schema.id
            self.merge_slot(slot, merged_slots)
            # Add default ranges
            if slot.range is None:
                # Inverses will be handled later on in the process
                if not slot.inverse:
                    slot.range = self.schema.default_range

        # Update enums
        for enum in self.schema.enums.values():
            if not enum.from_schema:
                enum.from_schema = self.schema.id
            # TODO: Need to add "is_a" to enums
            # self.merge_enum(enum, merged_enums)

        # Process the slot_usages
        for cls in self.schema.classes.values():
            self.process_slot_usages(cls)
            if not cls.from_schema:
                cls.from_schema = self.schema.id

        # Merge class with its mixins and the like
        merged_classes: list[ClassDefinitionName] = []
        for cls in self.schema.classes.values():
            self.merge_class(cls, merged_classes)

        # Update types with parental information
        merged_types: list[TypeDefinitionName] = []
        for typ in self.schema.types.values():
            if not typ.base and not typ.typeof:
                self.raise_value_error(
                    f'type "{typ.name}" must declare a type base or parent (typeof)',
                    typ.name,
                )
            if not typ.typeof and not typ.uri:
                self.raise_value_error(f'type "{typ.name}" does not declare a URI', typ.name)
            self.merge_type(typ, merged_types)
            if not typ.from_schema:
                typ.from_schema = self.schema.id

        # Update the subsets as needed
        for ss in self.schema.subsets.values():
            if not ss.from_schema:
                ss.from_schema = self.schema.id

        # Massage initial set of slots
        for slot in self.schema.slots.values():
            # Keys and identifiers must be present
            if bool(slot.key or slot.identifier):
                if slot.required is None:
                    slot.required = True
                elif not slot.required:
                    self.raise_value_error(
                        f"slot: {slot.name} - key and identifier slots cannot be optional",
                        slot.name,
                    )
                if slot.key and slot.identifier:
                    self.raise_value_error(
                        f"slot: {slot.name} - A slot cannot be both a key and identifier at the same time",
                        slot.name,
                    )

            # Propagate domain to containing class
            if slot.domain and slot.domain in self.schema.classes:
                if slot.name not in self.schema.classes[slot.domain].slots and not slot.owner:
                    slot.owner = slot.name
                    # Slot domains to not appear
                    # self.schema.classes[slot.domain].slots.append(slot.name)
            elif slot.domain:
                self.raise_value_error(
                    f"slot: {slot.name} - unrecognized domain ({slot.domain})",
                    slot.domain,
                )

            # Keys and identifiers must be present
            if bool(slot.key or slot.identifier):
                if slot.required is None:
                    slot.required = True
                elif not slot.required:
                    self.raise_value_error(
                        f"slot: {slot.name} - key and identifier slots cannot be optional",
                        slot.name,
                    )

            # Validate the slot range
            if (
                slot.range is not None
                and slot.range not in self.schema.types
                and slot.range not in self.schema.classes
                and slot.range not in self.schema.enums
            ):
                self.raise_value_error(f"slot: {slot.name} - unrecognized range ({slot.range})", slot.range)

            # check constraints for usage of equals_string and equals_string_in
            self._check_equals_string(slot)

        # Massage classes, propagating class slots entries domain back to the target slots
        for cls in self.schema.classes.values():
            if not isinstance(cls, ClassDefinition):
                name = cls["name"] if "name" in cls else "Unknown"
                self.raise_value_error(f'Class "{name} (type: {type(cls)})" definition is not a class definition')
            if isinstance(cls.slots, str):
                self.logger.warning(f"File: {self.schema.source_file} Class: {cls.name} Slots are not an array")
                cls.slots = [cls.slots]
            for slotname in cls.slots:
                if slotname in self.schema.slots:
                    slot = self.schema.slots[cast(SlotDefinitionName, slotname)]
                else:
                    self.raise_value_error(f'Class "{cls.name}" - unknown slot: "{slotname}"', slotname)

        for slot in self.schema.slots.values():
            if slot.from_schema is None:
                slot.from_schema = self.schema.id
            # Inline any class definitions that don't have identifiers.  Note that keys ARE inlined
            if slot.range in self.schema.classes:
                range_class = self.schema.classes[cast(ClassDefinitionName, slot.range)]
                range_has_identifier = any(
                    [self.schema.slots[s].identifier or self.schema.slots[s].key for s in range_class.slots]
                )
                # the inference and validation of class inlining is quite complex
                # slot.inlined and slot.inlined_as_list can have three different values:
                # True, False or None (if nothing specified in the schema)
                # range_has_identifier can only be True or False
                if range_has_identifier is True:
                    if slot.inlined is False:
                        if slot.inlined_as_list is True:
                            slot.inlined = True
                            # TODO: do not override a given value, but raise an exception
                            self.logger.warning(
                                f"Slot '{slot.name}': 'inlining' as list requested. "
                                + "Overriding inconsistent `inlined: False`!!"
                            )
                        elif slot.inlined_as_list is False:
                            self.logger.warning(
                                f"Slot '{slot.name}': Specifying a value for 'inlined_as_list' does not "
                                + "make any sense for non-inlined classes."
                            )
                    elif slot.inlined is None:
                        if slot.inlined_as_list is True:
                            slot.inlined = True
                            self.logger.info(
                                f"Slot '{slot.name}': A value for 'inlined_as_list' specified, but no 'inlined' "
                                + "specification. `inlined: True` has been automatically set!"
                            )
                        elif slot.inlined_as_list is False:
                            # TODO: decide if `inlined` should be set or not and implement accordingly
                            self.logger.warning(
                                f"Slot '{slot.name}': A value for 'inlined_as_list' specified, but no "
                                + "'inlined' specification."
                            )
                    elif slot.inlined is True:
                        if slot.inlined_as_list is None:
                            slot.inlined_as_list = False
                            self.logger.info(
                                f"Slot '{slot.name}': A class with identifiers ({range_class.name}) will be "
                                + "'inlined' as a dictionary. `inlined_as_list: False` has been set!"
                            )
                else:
                    if slot.inlined is True:
                        if slot.inlined_as_list is False:
                            self.logger.warning(
                                f"Slot '{slot.name}': A class without identifiers ({range_class.name}) cannot be "
                                + "'inlined' as a dictionary."
                            )
                        elif slot.inlined_as_list is None:
                            slot.inlined_as_list = True
                            self.logger.info(
                                f"Slot '{slot.name}': A class without identifiers ({range_class.name}) will not be "
                                + "typically 'inlined' as a dictionary. `inlined_as_list: True` has been set!"
                            )
                    elif slot.inlined is False:
                        slot.inlined = True
                        self.logger.info(
                            f"Slot '{slot.name}': A class without identifiers ({range_class.name}) MUST be "
                            + "'inlined', but non-inlined specified. `inlined: True` has been automatically set!"
                        )
                    else:
                        slot.inlined = True
                        self.logger.info(
                            f"Slot '{slot.name}': A class without identifiers ({range_class.name}) MUST be "
                            + "inlined, but nothing specified. `inlined: True` has been automatically set!"
                        )
                        if slot.inlined_as_list is False:
                            self.logger.warning(
                                f"Slot '{slot.name}': A class without identifiers ({range_class.name}) cannot be "
                                + "'inlined' as a dictionary."
                            )
                        elif slot.inlined_as_list is None:
                            slot.inlined_as_list = True
                            self.logger.info(
                                f"Slot '{slot.name}': A class without identifiers ({range_class.name}) will not be "
                                + "typically 'inlined' as a dictionary. `inlined_as_list: True` has been set!"
                            )

            if slot.slot_uri is not None:
                slot.mappings.insert(0, slot.slot_uri)
            # Assign missing predicates
            if slot.slot_uri is None or not self.useuris:
                slot.slot_uri = self.namespaces.uri_or_curie_for(
                    self.schema_defaults.get(slot.from_schema, sfx(slot.from_schema)),
                    self.slot_name_for(slot),
                )

            if slot.subproperty_of and slot.subproperty_of not in self.schema.slots:
                self.raise_value_error(
                    f'Slot: "{slot.name}" - subproperty_of: "{slot.subproperty_of}" '
                    f"does not reference a slot definition",
                    slot.subproperty_of,
                )

        # Evaluate any slot inverses
        def domain_range_alignment(fwd_slot: SlotDefinition, inverse_slot: SlotDefinition) -> bool:
            """Determine whether the range of fwd_slot is compatible with the domain of inverse_slot"""
            # TODO: Determine what to do about class and slot hierarchy
            if fwd_slot.range and fwd_slot.range not in self.schema.classes:
                raise ValueError(
                    f"Slot '{fwd_slot.name}' range ({fwd_slot.range}) is not an class -- inverse is not possible"
                )
            if fwd_slot.domain:
                if not inverse_slot.range:
                    inverse_slot.range = fwd_slot.domain
                elif not domain_range_alignment(fwd_slot, inverse_slot):
                    self.logger.warning(f"Slot: {slot.name} and inverse slot: {inverse_slot.name} are not compatible")
            return True

        # Get the inverse domains and ranges sorted
        for slot in self.schema.slots.values():
            if slot.inverse:
                # Note that the inverse OF the inverse will be caught in this same iterator
                inverse_slot = self.schema.slots[slot.inverse]
                if not slot.range:
                    if inverse_slot.domain:
                        slot.range = inverse_slot.domain
                    elif len(inverse_slot.domain_of):
                        if len(inverse_slot.domain_of) > 1:
                            dom_list = ", ".join(inverse_slot.domain_of)
                            self.logger.warning(
                                f"Slot {slot.name}.inverse ({inverse_slot.name}), "
                                f"has multi domains ({dom_list})  Multi ranges not yet implemented"
                            )
                        slot.range = inverse_slot.domain_of[0]
                    else:
                        raise ValueError(
                            f"Unable to determine the range of slot `{slot.name}'. "
                            f"Its inverse ({inverse_slot.name}) has no declared domain"
                        )
                elif not inverse_slot.domain and len(inverse_slot.domain_of) == 0:
                    inverse_slot.domain = slot.range
                elif slot.range not in (inverse_slot.domain, inverse_slot.domain_of):
                    self.logger.warning(
                        f"Range of slot '{slot.name}' ({slot.range}) "
                        f"does not line with the domain of its inverse ({inverse_slot.name})"
                    )

        # Check for duplicate class and type names
        def check_dups(s1: set[ElementName], s2: set[ElementName]) -> tuple[list[ElementName], str]:
            if s1.isdisjoint(s2):
                return [], ""

            # Return an ordered list of d1/d1 tuples
            # For some curious reason, s1.intersection(s2) and s2.intersection(s1) BOTH yield s1 elements
            dups = sorted(s1.intersection(s2))
            dup_locs = list()
            for dup in dups:
                dup_locs += [s1e for s1e in s1 if s1e == dup]
                dup_locs += [s2e for s2e in s2 if s2e == dup]

            return dup_locs, ", ".join(dups)

        classes = set(self.schema.classes.keys())
        self.validate_item_names("class", classes)
        slots = set(self.schema.slots.keys())
        self.validate_item_names("slot", slots)
        types = set(self.schema.types.keys())
        self.validate_item_names("type", types)
        subsets = set(self.schema.subsets.keys())
        self.validate_item_names("subset", subsets)
        enums = set(self.schema.enums.keys())
        self.validate_item_names("enum", enums)

        # Check that the default range is valid
        default_range_needed = any(slot.range == self.schema.default_range for slot in self.schema.slots.values())
        if (
            default_range_needed
            and self.schema.default_range not in self.schema.types
            and self.schema.default_range not in self.schema.classes
        ):
            raise ValueError(f'Unknown default range: "{self.schema.default_range}"')

        # We are currently limited to one key per class
        for cls in self.schema.classes.values():
            class_slots = []
            for sn in cls.slots:
                slot = self.schema.slots[sn]
                if slot.key or slot.identifier:
                    class_slots.append(sn)
            if len(class_slots) > 1:
                self.raise_value_error(
                    f'Class "{cls.name}" - multiple keys/identifiers not allowed ({", ".join(class_slots)})',
                    class_slots[1],
                )

        # Check out all the namespaces
        self.check_prefixes()

        # Cannot have duplicate class or type keys
        dups, items = check_dups(types, classes)
        if items:
            self.raise_value_errors(f"Overlapping type and class names: {items}", dups)
        dups, items = check_dups(enums, classes)
        if items:
            self.raise_value_errors(f"Overlapping enum and class names: {items}", dups)
        dups, items = check_dups(types, enums)
        if items:
            self.raise_value_errors(f"Overlapping type and enum names: {items}", dups)

        dups, items = check_dups(slots, classes)
        if items:
            self.logger_warning(f"Overlapping slot and class names: {items}", dups)

        dups, items = check_dups(subsets, classes)
        if items:
            self.logger_warning(f"Overlapping subset and class names: {items}", dups)

        dups, items = check_dups(types, slots)
        if items:
            self.logger_warning(f"Overlapping type and slot names: {items}", dups)

        dups, items = check_dups(subsets, slots)
        if items:
            self.logger_warning(f"Overlapping subset and slot names: {items}", dups)

        dups, items = check_dups(subsets, types)
        if items:
            self.logger_warning(f"Overlapping subset and type names: {items}", dups)

        dups, items = check_dups(enums, slots)
        if items:
            self.logger_warning(f"Overlapping enum and slot names: {items}", dups)

        dups, items = check_dups(subsets, enums)
        if items:
            self.logger_warning(f"Overlapping subset and enum names: {items}", dups)

        # Check over the various enumeration constraints
        for enum in self.schema.enums.values():
            if enum.code_set_version:
                if enum.code_set_tag:
                    self.raise_value_errors(
                        f'Enum: "{enum.name}" cannot have both version and tag',
                        [enum.code_set_version, enum.code_set_tag],
                    )
                if not enum.code_set:
                    self.raise_value_error(
                        f'Enum: "{enum.name}" needs a code set to have a version',
                        enum.name,
                    )
            if enum.code_set_tag:
                if not enum.code_set:
                    self.raise_value_error(f'Enum: "{enum.name}" needs a code set to have a tag', enum.name)
            if enum.pv_formula:
                if not enum.code_set:
                    self.raise_value_error(
                        f'Enum: "{enum.name}" needs a code set to have a formula',
                        enum.name,
                    )
                if enum.permissible_values:
                    self.raise_value_error(
                        f'Enum: "{enum.name}" can have a formula or permissible values but not both',
                        enum.name,
                    )
        for slot in self.schema.slots.values():
            if slot.range and slot.range in self.schema.enums:
                if slot.inlined or slot.inlined_as_list:
                    self.raise_value_error(
                        f'Slot: "{slot.name}" enumerations cannot be inlined',
                        slot.range,
                    )

        # Make the source file relative if it is locally generated
        self.schema_location = self.schema.source_file
        if self.schema.source_file and "://" not in self.schema.source_file:
            self.schema.source_file = os.path.basename(self.schema.source_file)

        # Make sure there is only one tree_root
        tree_root = None
        for cls in self.schema.classes.values():
            if cls.tree_root:
                if tree_root is not None:
                    self.logger.warning(f"Duplicate tree_root: {cls.name} with {tree_root}")
                else:
                    tree_root = cls.name

        self.synopsis = SchemaSynopsis(self.schema)
        errs = self.synopsis.errors()
        if errs:
            print("Warning: The following errors were encountered in the schema")
            for errline in errs:
                print("\t" + errline)
            print()
        for subset, referees in self.synopsis.subsetrefs.items():
            if subset not in self.schema.subsets:
                self.raise_value_error(f"Subset: {subset} is not defined", subset)
        return self.schema

    def validate_item_names(self, typ: str, names: list[str]) -> None:
        # TODO: add a more rigorous syntax check for item names
        for name in names:
            if ":" in name:
                raise self.raise_value_error(f'{typ}: "{name}" - ":" not allowed in identifier', name)

    def merge_enum(self, enum: EnumDefinition, merged_enums: list[EnumDefinitionName]) -> None:
        """
        Merge parent enumeration information into target enum

        :param enum: target enumeration
        :param merged_enums: list of enum names that have been merged. Used to do distal ancestor resolution
        """
        if enum.name not in merged_enums:
            merged_enums.append(enum.name)
            if enum.is_a:
                if enum.is_a in self.schema.enums:
                    self.merge_enum(self.schema.enums[enum.is_a], merged_enums)
                    # merge_enums(self.schema, enum, self.schema.enums[enum.is_a], False)
                else:
                    self.raise_value_error(
                        f'Enum: "{enum.name}" - unknown is_a reference: {enum.is_a}',
                        enum.is_a,
                    )

    def merge_slot(self, slot: SlotDefinition, merged_slots: list[SlotDefinitionName]) -> None:
        """
        Merge parent slot information into target slot

        :param slot: target slot
        :param merged_slots: list of slot names that have been merged.  Used to do a distal ancestor resolution
        """
        if slot.name not in merged_slots:
            if slot.is_a:
                try:
                    if slot.is_a in self.schema.slots:
                        self.merge_slot(self.schema.slots[slot.is_a], merged_slots)
                        merge_slots(slot, self.schema.slots[slot.is_a])
                    else:
                        self.raise_value_error(
                            f'Slot: "{slot.name}" - unknown is_a reference: {slot.is_a}',
                            slot.is_a,
                        )
                except RecursionError:
                    self.raise_value_error(
                        f'Slot: "{slot.name}" - recursive is_a reference: {slot.is_a}',
                        slot.is_a,
                    )

            for mixin in slot.mixins:
                if mixin in self.schema.slots:
                    self.merge_slot(self.schema.slots[mixin], merged_slots)
                    merge_slots(slot, self.schema.slots[mixin])
                else:
                    self.raise_value_error(f'Slot: "{slot.name}" - unknown mixin reference: {mixin}', mixin)
            merged_slots.append(slot.name)

    def merge_class(self, cls: ClassDefinition, merged_classes: list[ClassDefinitionName]) -> None:
        """
        Merge parent class information into target class

        :param cls: target class
        :param merged_classes: list of class names that have been merged. Used to do distal ancestor resolution
        """
        if cls.name not in merged_classes:
            merged_classes.append(cls.name)
            if cls.is_a:
                if cls.is_a in self.schema.classes:
                    self.merge_class(self.schema.classes[cls.is_a], merged_classes)
                    merge_classes(self.schema, cls, self.schema.classes[cls.is_a], False)
                else:
                    self.raise_value_error(
                        f'Class: "{cls.name}" - unknown is_a reference: {cls.is_a}',
                        cls.is_a,
                    )
            for mixin in cls.mixins:
                # Note that apply_to has been injected as a faux mixin, so it gets covered here
                if mixin in self.schema.classes:
                    self.merge_class(self.schema.classes[mixin], merged_classes)
                    merge_classes(self.schema, cls, self.schema.classes[mixin], True)
                else:
                    self.raise_value_error(f'Class: "{cls.name}" - unknown mixin reference: {mixin}', mixin)

    def process_slot_usage_definitions(self):
        """
        Slot usages can be used to completely define slots.  Iterate over the class hierarchy finding all slot
        definitions that are introduced strictly as usages and add them to the slots component
        """
        visited: set[ClassDefinitionName] = set()
        visited_usages: set[SlotDefinitionName] = set()  # Slots that are or will be mangled

        def located_aliased_parent_slot(owning_class: ClassDefinition, usage_slot: SlotDefinition) -> bool:
            """Determine whether we are overriding an attributes style slot in the parent class
            Preconditions: usage_slot is NOT in schema.slots
            """
            usage_attribute_name = mangled_attribute_name(owning_class.name, usage_slot.name)
            if owning_class.is_a:
                parent_slot_name = mangled_attribute_name(owning_class.is_a, usage_slot.name)
                if parent_slot_name in self.schema.slots or parent_slot_name in visited_usages:
                    usage_slot.is_a = parent_slot_name
                    visited_usages.add(usage_attribute_name)
                    return True
            for mixin in owning_class.mixins:
                mixin_slot_name = mangled_attribute_name(mixin, usage_slot.name)
                if mixin_slot_name in self.schema.slots or mixin_slot_name in visited_usages:
                    usage_slot.is_a = mixin_slot_name
                    visited_usages.add(usage_attribute_name)
                    return True
            return False

        def visit(classname: ClassDefinitionName) -> None:
            cls = self.schema.classes.get(classname)
            if cls and cls.name not in visited:
                if cls.is_a:
                    visit(cls.is_a)
                for mixin in cls.mixins:
                    visit(mixin)
                for slot_usage in values(cls.slot_usage):
                    if slot_usage.alias:
                        self.raise_value_error(
                            f'Class: "{cls.name}" - alias not permitted in slot_usage slot: {slot_usage.alias}'
                        )
                    if not located_aliased_parent_slot(cls, slot_usage):
                        if slot_usage.name not in self.schema.slots:
                            self.logger.info(
                                f'class "{cls.name}" slot "{slot_usage.name}" '
                                f"does not reference an existing slot.  New slot was created."
                            )
                            # TODO: Consider tightening this up and only allowing usages on defined slots
                            self.schema.slots[slot_usage.name] = slot_usage
                        else:
                            # TODO Make sure that the slot_usage.name is legal (occurs in an ancestor of the class
                            pass
                visited.add(classname)

        for classname in self.schema.classes.keys():
            visit(classname)

    def process_slot_usages(self, cls: ClassDefinition) -> None:
        """
        Connect any slot usage items

        :param cls: class to process
        :return: usage item
        """
        for slotname, slot_usage in cls.slot_usage.items():
            if slot_usage.alias:
                self.raise_value_error(
                    f'Class: "{cls.name}" - alias not permitted in slot_usage slot: {slot_usage.alias}'
                )
            # Construct a new slot
            # If we've already assigned a parent, use it

            if slotname in self.schema.slots:
                base_slot = self.schema.slots[slotname]
            else:
                self.logger.error(f"slot_usage for undefined slot: {slotname}")
                base_slot = None
            parent_slot = self.schema.slots.get(slot_usage.is_a)
            # Follow the ancestry of the class to get the most proximal parent
            if not parent_slot:
                parent_slot = self.slot_definition_for(slotname, cls)
            if not parent_slot and slotname in self.schema.slots:
                parent_slot = self.schema.slots[slotname]

            if not parent_slot:
                # This test is here because it is really easy to break things in the slot merge utilities.  It should
                # stay
                self.logger.error(f'class "{cls.name}" slot "{slotname}" -- error occurred. This should not happen')
            else:
                child_name = slot_usage_name(slotname, cls)
                slot_alias = parent_slot.alias if parent_slot.alias else slotname
            new_slot = SlotDefinition(
                name=child_name,
                alias=slot_alias,
                domain=cls.name,
                is_usage_slot=Bool(True),
                usage_slot_name=slotname,
                owner=cls.name,
                domain_of=[cls.name],
                imported_from=cls.imported_from,
            )

            self.schema.slots[child_name] = new_slot
            merge_slots(
                new_slot,
                slot_usage,
                inheriting=False,
                skip=[
                    "name",
                    "alias",
                    "domain",
                    "is_usage_slot",
                    "usage_slot_name",
                    "owner",
                    "domain_of",
                ],
            )
            # Copy the parent definition.  If there is no parent definition, the slot is being defined
            # locally as a slot_usage
            if parent_slot is not None:
                new_slot.is_a = parent_slot.name
                merge_slots(new_slot, parent_slot)
                # This situation occurs when we are doing chained overrides.  Kludgy, but it works...
                if parent_slot.name in cls.slots:
                    if child_name in cls.slots:
                        del cls.slots[cls.slots.index(child_name)]
                    cls.slots[cls.slots.index(parent_slot.name)] = child_name
                elif child_name not in cls.slots:
                    cls.slots.append(child_name)
            elif not new_slot.range:
                new_slot.range = self.schema.default_range
            # copy base slot metalsot values across, except where already
            # populated/overridden, OR where propagation to induced slots is
            # forbidden (inverses)
            if base_slot is not None:
                for metaslot_name in base_slot.__dict__.keys():
                    current_val = getattr(new_slot, metaslot_name)
                    if not current_val and metaslot_name not in ["inverse"]:
                        new_val = deepcopy(getattr(base_slot, metaslot_name))
                        if new_val:
                            setattr(new_slot, metaslot_name, new_val)

    def merge_type(self, typ: TypeDefinition, merged_types: list[TypeDefinitionName]) -> None:
        """
        Merge parent type information into target type
        :param typ: target type
        :param merged_types: list of type names that have bee merged.
        """
        if typ.name not in merged_types:
            if typ.typeof:
                if typ.typeof in self.schema.types:
                    reftyp = self.schema.types[cast(TypeDefinitionName, typ.typeof)]
                    self.merge_type(reftyp, merged_types)
                    merge_slots(typ, reftyp, [SlotDefinitionName("imported_from")])
                else:
                    self.raise_value_error(
                        f'Type: "{typ.name}" - unknown typeof reference: {typ.typeof}',
                        typ.typeof,
                    )
            merged_types.append(typ.name)

    def schema_errors(self) -> list[str]:
        return self.synopsis.errors() if self.synopsis else ["resolve() must be run before error check"]

    def slot_definition_for(self, slotname: SlotDefinitionName, cls: ClassDefinition) -> Optional[SlotDefinition]:
        """Find the most proximal definition for slotname in the context of cls"""
        if cls.is_a:
            if cls.is_a not in self.schema.classes:
                self.raise_value_error(f"Unknown parent class: {cls.is_a}", cls.is_a)
            for sn in self.schema.classes[cls.is_a].slots:
                slot = self.schema.slots[sn]
                if (slot.usage_slot_name and slotname == slot.usage_slot_name) or (
                    not slot.usage_slot_name and slotname == slot.name
                ):
                    return slot
        for mixin in cls.mixins:
            if mixin not in self.schema.classes:
                self.raise_value_error(f"Unknown mixin class: {mixin}", cls.is_a)
            for sn in self.schema.classes[mixin].slots:
                slot = self.schema.slots[sn]
                if slot.alias and slotname == slot.alias or slotname == slot.name:
                    return slot
        if cls.is_a:
            defn = self.slot_definition_for(slotname, self.schema.classes[cls.is_a])
            if defn:
                return defn
        for mixin in cls.mixins:
            defn = self.slot_definition_for(slotname, self.schema.classes[mixin])
            if defn:
                return defn
        return None

    def check_prefixes(self) -> None:
        """
        Iterate over the entire schema checking all prefixes
        """
        self.check_prefix(self.schema.default_prefix)
        for prefix in self.schema.emit_prefixes:
            self.check_prefix(prefix)
        for typ in self.schema.types.values():
            self.check_prefix(typ.uri)
            for prefix in typ.mappings:
                self.check_prefix(prefix)
            for prefix in typ.id_prefixes:
                self.check_prefix(prefix)
        for slot in self.schema.slots.values():
            self.check_prefix(slot.slot_uri)
            for prefix in slot.mappings:
                self.check_prefix(prefix)
            for prefix in slot.id_prefixes:
                self.check_prefix(prefix)
        for cls in self.schema.classes.values():
            self.check_prefix(cls.class_uri)
            # Class URI's are inserted into mappings -- see line ~#184
            for prefix in cls.mappings:
                if prefix != cls.class_uri:
                    self.check_prefix(prefix)
            for prefix in cls.id_prefixes:
                self.check_prefix(prefix)

    def check_prefix(self, prefix_or_curie_or_uri: str) -> None:
        prefix = self.namespaces.prefix_for(prefix_or_curie_or_uri, case_shift=False)
        if prefix:
            if prefix not in self.namespaces:
                self.logger.warning(f"{TypedNode.yaml_loc(prefix_or_curie_or_uri)}Unrecognized prefix: {prefix}")
                self.namespaces[prefix] = f"http://example.org/UNKNOWN/{prefix}/"
            else:
                case_adjusted_prefix = self.namespaces.prefix_for(prefix_or_curie_or_uri, case_shift=True)
                if case_adjusted_prefix != prefix:
                    self.logger.warning(
                        f"{TypedNode.yaml_loc(prefix_or_curie_or_uri)}"
                        f"Prefix case mismatch - supplied: {prefix} "
                        f"expected: {case_adjusted_prefix}"
                    )

    @staticmethod
    def slot_name_for(slot: SlotDefinition) -> str:
        return underscore(slot.alias if slot.alias else slot.name)

    @staticmethod
    def raise_value_error(error: str, loc_str: Optional[Union[TypedNode, str]] = None) -> None:
        SchemaLoader.raise_value_errors(error, loc_str)

    @staticmethod
    def raise_value_errors(error: str, loc_str: Optional[Union[str, TypedNode, Iterator[TypedNode]]]) -> None:
        if isinstance(loc_str, list):
            locs = "\n".join(TypedNode.yaml_loc(e, suffix="") for e in loc_str)
            raise ValueError(f"{locs} {error}")
        else:
            raise ValueError(f"{TypedNode.yaml_loc(loc_str, suffix='')} {error}")

    def logger_warning(
        self,
        warning: str,
        loc_str: Optional[Union[str, TypedNode, Iterator[TypedNode]]],
    ) -> None:
        if isinstance(loc_str, list):
            locs = "\n\t".join(TypedNode.yaml_loc(e, suffix="") for e in loc_str)
            self.logger.warning(f"{warning}\n\t{locs}")
        else:
            self.logger.warning(f"{warning}\n\t{TypedNode.yaml_loc(loc_str, suffix='')}")

    def _get_base_dir(self, stated_base: str) -> Optional[str]:
        if stated_base:
            return stated_base
        elif self.schema.source_file:
            if "://" in self.schema.source_file:
                parsed_url = urlparse(self.schema.source_file)
                self.schema.source_file = parsed_url.path.rsplit("/", 1)[-1]
                return parsed_url.path.split("/", 1)[0]
            else:
                rval = os.path.dirname(os.path.abspath(self.schema.source_file))
                return rval
        else:
            return None

    def _check_equals_string(self, slot: SlotDefinition):
        if slot.equals_string or slot.equals_string_in:
            # Range "string" mandatory for "equals_string" and "equals_string_in"
            range = slot.range
            if not range:
                # range is not defined --> check default range
                range = self.schema.default_range
            if range != "string":
                self.raise_value_error(
                    f"slot: {slot.name} - 'equals_string' and 'equals_string_in' requires range "
                    f"'string' and not range '{range}'",
                    slot.range,
                )
            if slot.any_of:
                # It is not allowed to use any of and equals_string or equals_string_in in one slot definition,
                # as both are mapped to sh:in in SHACL
                self.raise_value_error(
                    f"slot: {slot.name} - 'equals_string'/'equals_string_in' and 'any_of' are mutually exclusive",
                    slot.name,
                )



import os
from dataclasses import dataclass
from typing import Optional

import click
from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition
from linkml_runtime.utils.formatutils import camelcase, lcamelcase

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class ProtoGenerator(Generator):
    """
    A `Generator` for creating Protobuf schemas from a linkml schema.

    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["proto"]
    visit_all_class_slots = True
    uses_schemaloader = True

    # ObjectVars
    relative_slot_num: int = 0

    def visit_schema(self, **kwargs) -> Optional[str]:
        return self.generate_header()

    def generate_header(self) -> str:
        items = []
        items.append(' syntax="proto3";')
        items.append(" package")
        items.append(f"// metamodel_version: {self.schema.metamodel_version}")
        if self.schema.version:
            items.append(f"// version: {self.schema.version}")
        out = "\n".join(items) + "\n"
        return out

    def visit_class(self, cls: ClassDefinition) -> Optional[str]:
        if cls.mixin or cls.abstract or not cls.slots:
            return None
        items = []
        if cls.description:
            for dline in cls.description.split("\n"):
                items.append(f"// {dline}")
        items.append(f"message {camelcase(cls.name)}")
        items.append(" {")
        out = "\n".join(items)
        self.relative_slot_num = 0
        return out

    def end_class(self, cls: ClassDefinition) -> str:
        return "\n }\n"

    def visit_class_slot(self, cls: ClassDefinition, aliased_slot_name: str, slot: SlotDefinition) -> str:
        qual = "repeated " if slot.multivalued else ""
        slotname = lcamelcase(aliased_slot_name)
        slot_range = camelcase(slot.range)
        if slot.rank is None:
            # numbering of slots is important in the proto implementation
            # and should be determined by the rank param.
            slot.rank = 0
        return f"\n {qual} {lcamelcase(slot_range)} {(slotname)} = {slot.rank}"


@shared_arguments(ProtoGenerator)
@click.version_option(__version__, "-V", "--version")
@click.command(name="proto")
def cli(yamlfile, **args):
    """Generate proto representation of LinkML model"""
    print(ProtoGenerator(yamlfile, **args).serialize(**args))


if __name__ == "__main__":
    cli()



import keyword
import logging
import os
import re
from collections.abc import Iterator
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Callable, Optional, Union

import click
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import linkml_files
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    DefinitionName,
    Element,
    EnumDefinition,
    PermissibleValue,
    SlotDefinition,
    SlotDefinitionName,
    TypeDefinition,
)
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.formatutils import be, camelcase, sfx, split_col, underscore, wrapped_annotation
from linkml_runtime.utils.metamodelcore import builtinnames
from rdflib import URIRef

import linkml
from linkml._version import __version__
from linkml.generators.pydanticgen.template import Import, Imports, ObjectImport
from linkml.generators.python.python_ifabsent_processor import PythonIfAbsentProcessor
from linkml.utils.generator import Generator, shared_arguments

logger = logging.getLogger(__name__)


@dataclass
class PythonGenerator(Generator):
    """
    Generates Python dataclasses from a LinkML model

    See `Python Generator Docs <https://linkml.io/linkml/generators/python.html>`_
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["py"]
    file_extension = "py"
    visit_all_class_slots = False
    uses_schemaloader = True

    # ObjectVars
    gen_classvars: bool = True
    gen_slots: bool = True
    genmeta: bool = False
    emit_metadata: bool = True
    dataclass_repr: bool = False
    """
    Whether generated dataclasses should also generate a default __repr__ method.

    Default ``False`` so that the parent :class:`linkml_runtime.utils.yamlutils.YAMLRoot` 's
    ``__repr__`` method is inherited for model pretty printing.

    References:
        - https://docs.python.org/3/library/dataclasses.html#dataclasses.dataclass
    """

    def __post_init__(self) -> None:
        if isinstance(self.schema, Path):
            self.schema = str(self.schema)
        self.sourcefile = self.schema
        self.schemaview = SchemaView(self.schema, base_dir=self.base_dir)
        self.ifabsent_processor = PythonIfAbsentProcessor(self.schemaview)
        super().__post_init__()
        if self.format is None:
            self.format = self.valid_formats[0]
        if self.schema.default_prefix == "linkml" and not self.genmeta:
            logger.error("Generating metamodel without --genmeta is highly inadvisable!")
        if not self.schema.source_file and isinstance(self.sourcefile, str) and "\n" not in self.sourcefile:
            self.schema.source_file = os.path.basename(self.sourcefile)

    def compile_module(self, **kwargs) -> ModuleType:
        """
        Compiles generated python code to a module
        :return:
        """
        pycode = self.serialize(**kwargs)
        try:
            return compile_python(pycode)
        except NameError as e:
            logger.error(f"Code:\n{pycode}")
            logger.error(f"Error compiling generated python code: {e}")
            raise e

    def visit_schema(self, **kwargs) -> None:
        # Add explicitly declared prefixes
        self.emit_prefixes.update([p.prefix_prefix for p in self.schema.prefixes.values()])

        # Add all emit statements
        self.emit_prefixes.update(self.schema.emit_prefixes)

        # Add the default prefix
        if self.schema.default_prefix:
            self.emit_prefixes.add(self.namespaces.prefix_for(self.schema.default_prefix))

    def visit_class(self, cls: ClassDefinition) -> bool:
        if not cls.imported_from:
            cls_prefix = self.namespaces.prefix_for(cls.class_uri)
            if cls_prefix:
                self.emit_prefixes.add(cls_prefix)
            self.add_mappings(cls)
        return False

    def visit_slot(self, aliased_slot_name: str, slot: SlotDefinition) -> None:
        if not slot.imported_from:
            slot_prefix = self.namespaces.prefix_for(slot.slot_uri)
            if slot_prefix:
                self.emit_prefixes.add(slot_prefix)
            self.add_mappings(slot)

    def visit_type(self, typ: TypeDefinition) -> None:
        if not typ.imported_from:
            type_prefix = self.namespaces.prefix_for(typ.uri)
            if type_prefix:
                self.emit_prefixes.add(type_prefix)

    def gen_schema(self) -> str:
        all_imports = Imports()
        # generic imports
        all_imports = (
            all_imports
            + Import(module="dataclasses")
            + Import(module="re")
            + Import(
                module="jsonasobj2",
                objects=[
                    ObjectImport(name="JsonObj"),
                    ObjectImport(name="as_dict"),
                ],
            )
            + Import(
                module="typing",
                objects=[
                    ObjectImport(name="Optional"),
                    ObjectImport(name="List"),
                    ObjectImport(name="Union"),
                    ObjectImport(name="Dict"),
                    ObjectImport(name="ClassVar"),
                    ObjectImport(name="Any"),
                ],
            )
            + Import(
                module="dataclasses",
                objects=[
                    ObjectImport(name="dataclass"),
                ],
            )
            + Import(
                module="datetime",
                objects=[
                    ObjectImport(name="date"),
                    ObjectImport(name="datetime"),
                    ObjectImport(name="time"),
                ],
            )
        )

        # The metamodel uses Enumerations to define itself, so don't import if we are generating the metamodel
        if not self.genmeta:
            all_imports = all_imports + Import(
                module="linkml_runtime.linkml_model.meta",
                objects=[
                    ObjectImport(name="EnumDefinition"),
                    ObjectImport(name="PermissibleValue"),
                    ObjectImport(name="PvFormulaOptions"),
                ],
            )
        # linkml imports
        all_imports = (
            all_imports
            + Import(
                module="linkml_runtime.utils.slot",
                objects=[
                    ObjectImport(name="Slot"),
                ],
            )
            + Import(
                module="linkml_runtime.utils.metamodelcore",
                objects=[
                    ObjectImport(name="empty_list"),
                    ObjectImport(name="empty_dict"),
                    ObjectImport(name="bnode"),
                ],
            )
            + Import(
                module="linkml_runtime.utils.yamlutils",
                objects=[
                    ObjectImport(name="YAMLRoot"),
                    ObjectImport(name="extended_str"),
                    ObjectImport(name="extended_float"),
                    ObjectImport(name="extended_int"),
                ],
            )
            + Import(
                module="linkml_runtime.utils.formatutils",
                objects=[
                    ObjectImport(name="camelcase"),
                    ObjectImport(name="underscore"),
                    ObjectImport(name="sfx"),
                ],
            )
        )

        # handler import
        all_imports = all_imports + Import(
            module="linkml_runtime.utils.enumerations", objects=[ObjectImport(name="EnumDefinitionImpl")]
        )
        # other imports
        all_imports = (
            all_imports
            + Import(
                module="rdflib",
                objects=[
                    ObjectImport(name="Namespace"),
                    ObjectImport(name="URIRef"),
                ],
            )
            + Import(
                module="linkml_runtime.utils.curienamespace",
                objects=[
                    ObjectImport(name="CurieNamespace"),
                ],
            )
        )

        split_description = ""
        if self.schema.description:
            split_description = "\n#   ".join(d for d in self.schema.description.split("\n") if d is not None)
        head = (
            f"""# Auto generated from {self.schema.source_file} by {self.generatorname} version: {self.generatorversion}
# Generation date: {self.schema.generation_date}
# Schema: {self.schema.name}
#"""
            if self.emit_metadata and self.schema.generation_date
            else ""
        )

        return f"""{head}
# id: {self.schema.id}
# description: {split_description}
# license: {be(self.schema.license)}

{all_imports.render()}
{self.gen_imports()}

metamodel_version = "{self.schema.metamodel_version}"
version = {'"' + self.schema.version + '"' if self.schema.version else None}

# Namespaces
{self.gen_namespaces()}


# Types
{self.gen_typedefs()}
# Class references
{self.gen_references()}

{self.gen_classdefs()}

# Enumerations
{self.gen_enumerations()}

# Slots
{self.gen_slotdefs()}"""

    def end_schema(self, **_) -> str:
        return re.sub(r" +\n", "\n", self.gen_schema().replace("\t", "    ")).strip(" ")

    def gen_imports(self) -> str:
        list_ents = [f"from {k} import {', '.join(v)}" for k, v in self.gen_import_list().items()]
        return "\n".join(list_ents)

    def gen_import_list(self) -> dict[str, list[str]]:
        """
        Generate a list of types to import

        :return: source file followed by elements to import
        """

        class ImportList:
            def __init__(self, schema_location: str):
                self.schema_location = schema_location
                self.v: dict[str, set[str]] = {}

            def add_element(self, e: Element) -> None:
                if e.imported_from:
                    self.add_entry(e.imported_from, camelcase(e.name))

            def add_entry(innerself, path: Union[str, URIRef], name: str) -> None:
                path = str(self.namespaces.uri_for(path) if ":" in path else path)
                if path.startswith(linkml_files.LINKML_NAMESPACE):
                    model_base = "." if self.genmeta else "linkml_runtime.linkml_model."
                    innerself.v.setdefault(model_base + path[len(linkml_files.LINKML_NAMESPACE) :], set()).add(name)
                elif path == linkml.BIOLINK_MODEL_URI:
                    innerself.v.setdefault(linkml.BIOLINK_MODEL_PYTHON_LOC, set()).add(name)
                elif "://" in path:
                    raise ValueError(f"Cannot map {path} into a python import statement")
                elif "/" in path:
                    innerself.v.setdefault(path.replace("./", ".").replace("/", "."), set()).add(name)
                elif "." in path:
                    innerself.v.setdefault(path, set()).add(name)
                else:
                    innerself.v.setdefault(". " + path, set()).add(name)

            def values(self) -> dict[str, list[str]]:
                return {k: sorted(self.v[k]) for k in sorted(self.v.keys())}

        def add_type_ref(typ: TypeDefinition) -> None:
            if not typ.typeof and typ.base and typ.base not in builtinnames:
                if "." in typ.base:
                    rval.add_entry(*typ.base.rsplit("."))
                else:
                    rval.add_entry("linkml_runtime.utils.metamodelcore", typ.base)
            if typ.typeof:
                add_type_ref(self.schema.types[typ.typeof])
            rval.add_element(typ)

        def add_enum_ref(e: EnumDefinition) -> None:
            rval.add_element(e)

        def add_slot_range(slot: SlotDefinition) -> None:
            if slot.range:
                if slot.range in self.schema.types:
                    add_type_ref(self.schema.types[slot.range])
                elif slot.range in self.schema.enums:
                    add_enum_ref(self.schema.enums[slot.range])
                else:
                    cls = self.schema.classes[slot.range]
                    if cls.imported_from:
                        if self.class_identifier(cls):
                            identifier_range = self.class_identifier_path(cls, False)[-1]
                            if identifier_range in self.schema.types:
                                add_type_ref(TypeDefinition(identifier_range))
                            else:
                                rval.add_entry(cls.imported_from, identifier_range)
                        if slot.inlined:
                            rval.add_element(cls)

        rval = ImportList(self.schema_location)
        for typ in self.schema.types.values():
            if not typ.imported_from:
                add_type_ref(typ)
        for slot in self.schema.slots.values():
            if not slot.imported_from:
                if slot.is_a:
                    parent = self.schema.slots[slot.is_a]
                    if (parent.key or parent.identifier) and parent.imported_from:
                        rval.add_element(self.schema.slots[slot.is_a])
                if slot.domain:
                    domain = self.schema.classes[slot.domain]
                    if domain.imported_from:
                        rval.add_element(self.schema.classes[slot.domain])
                add_slot_range(slot)

        for cls in self.schema.classes.values():
            if not cls.imported_from:
                if cls.is_a:
                    parent = self.schema.classes[cls.is_a]
                    if parent.imported_from:
                        rval.add_element(self.schema.classes[cls.is_a])
                        if self.class_identifier(parent):
                            rval.add_entry(
                                parent.imported_from,
                                self.class_identifier_path(parent, False)[-1],
                            )
                for slotname in cls.slots:
                    add_slot_range(self.schema.slots[slotname])

        return rval.values()

    def gen_namespaces(self) -> str:
        dflt_prefix = self._default_curie_or_uri()
        dflt = f"CurieNamespace('', '{sfx(dflt_prefix)}')" if ":/" in dflt_prefix else dflt_prefix.upper()
        curienamespace_defs = [
            {
                "variable": f"{pfx.upper().replace('.', '_').replace('-', '_')}",
                "value": f"CurieNamespace('{pfx.replace('.', '_')}', '{self.namespaces[pfx]}')",
            }
            for pfx in sorted(self.emit_prefixes)
            if pfx in self.namespaces
        ]
        curienamespace_declarations = "\n".join(
            [f"{ns['variable']} = {ns['value']}" for ns in curienamespace_defs] + [f"DEFAULT_ = {dflt}"]
        )

        ",".join([x["variable"] for x in curienamespace_defs])
        # catalog_declaration = f"\nnamespace_catalog = CurieNamespaceCatalog.create({curienamespace_vars})\n"
        catalog_declaration = ""

        return curienamespace_declarations + catalog_declaration

    def gen_references(self) -> str:
        """Generate python type declarations for all identifiers (primary keys)"""
        rval = []
        for cls in self._sort_classes(self.schema.classes.values()):
            if not cls.imported_from:
                pkeys = self.primary_keys_for(cls)
                if pkeys:
                    for pk in pkeys:
                        classname = camelcase(cls.name) + camelcase(self.aliased_slot_name(pk))
                        # If we've got a parent slot and the range of the parent is the range of the child, the
                        # child slot is a subclass of the parent.  Otherwise, the child range has been overridden,
                        # so the inheritance chain has been broken
                        parent_pk = self.class_identifier(cls.is_a) if cls.is_a else None
                        parent_pk_slot = self.schema.slots[parent_pk] if parent_pk else None
                        pk_slot = self.schema.slots[pk]
                        if parent_pk_slot and (parent_pk_slot.name == pk or pk_slot.range == parent_pk_slot.range):
                            parents = self.class_identifier_path(cls.is_a, False)
                        else:
                            parents = self.slot_range_path(pk_slot)
                        parent_cls = (
                            "extended_" + parents[-1] if parents[-1] in ["str", "float", "int"] else parents[-1]
                        )
                        rval.append(f"class {classname}({parent_cls}):\n\tpass")
                        break  # We only do the first primary key
        return "\n\n\n".join(rval)

    def gen_typedefs(self) -> str:
        """Generate python type declarations for all defined types"""
        rval = []
        defs_to_generate = [x for x in self.schema.types.values() if not x.imported_from]
        emitted_types = []
        # all imported_from types are already considered generated
        emitted_types.extend([x.name for x in self.schema.types.values() if x.imported_from])
        for typ in [x for x in defs_to_generate if not x.typeof]:
            self._gen_typedef(typ, typ.base.rsplit(".")[-1], rval, emitted_types)

        while True:
            defs_to_generate_typeof = [x for x in defs_to_generate if x.typeof and x.name not in emitted_types]
            if len(defs_to_generate_typeof) == 0:
                break
            defs_can_generate = [x for x in defs_to_generate_typeof if x.typeof in emitted_types]
            if len(defs_can_generate) == 0:
                raise ValueError(
                    "Cannot generate type definition for "
                    f"{[f'{x.name} of {x.typeof}' for x in defs_to_generate_typeof]}. "
                    "Forgot a link in the type hierarchy chain?"
                )
            for typ in defs_can_generate:
                self._gen_typedef(typ, camelcase(typ.typeof), rval, emitted_types)

        return "\n".join(rval)

    def _gen_typedef(self, typ, superclass, rval, emitted_types):
        typname = camelcase(typ.name)
        desc = ""
        if typ.description:
            description = typ.description.replace('"""', "---")
            desc = f'\n\t""" {description} """'
        rval.append(f"class {typname}({superclass}):{desc}\n\t{self.gen_type_meta(typ)}\n\n")
        emitted_types.append(typ.name)

    def gen_classdefs(self) -> str:
        """Create class definitions for all non-mixin classes in the model
        Note that apply_to classes are transformed to mixins
        """
        clist = self._sort_classes(self.schema.classes.values())
        return "\n".join([self.gen_classdef(v) for v in clist if not v.imported_from])

    def gen_classdef(self, cls: ClassDefinition) -> str:
        """Generate python definition for class cls"""

        parentref = f"({self.formatted_element_name(cls.is_a, True) if cls.is_a else 'YAMLRoot'})"
        slotdefs = self.gen_class_variables(cls)
        postinits = self.gen_postinits(cls)
        constructor = self.gen_constructor(cls)

        wrapped_description = (
            f'\n\t"""\n\t{wrapped_annotation(be(cls.description))}\n\t"""' if be(cls.description) else ""
        )

        if self.is_class_unconstrained(cls):
            return f"\n{self.class_or_type_name(cls.name)} = Any"

        cd_str = (
            (f"\n@dataclass(repr={self.dataclass_repr})" if slotdefs else "")
            + f"\nclass {self.class_or_type_name(cls.name)}{parentref}:{wrapped_description}"
            + f"{self.gen_inherited_slots(cls)}"
            + f"{self.gen_class_meta(cls)}"
            + (f"\n\t{slotdefs}" if slotdefs else "")
            + (f"\n{postinits}" if postinits else "")
            + (f"\n{constructor}" if constructor else "")
        )

        return cd_str

    def gen_inherited_slots(self, cls: ClassDefinition) -> str:
        if not self.gen_classvars:
            return ""
        inherited_slots = []
        for slotname in cls.slots:
            slot = self.schema.slots[slotname]
            if slot.inherited:
                inherited_slots.append(slot.alias if slot.alias else slotname)
        inherited_slots_str = ", ".join([f'"{underscore(s)}"' for s in inherited_slots])
        return f"\n\t_inherited_slots: ClassVar[list[str]] = [{inherited_slots_str}]\n"

    def gen_class_meta(self, cls: ClassDefinition) -> str:
        if not self.gen_classvars:
            return ""
        class_class_uri = self.namespaces.uri_for(cls.class_uri)
        if class_class_uri:
            cls_python_uri = self.namespaces.curie_for(class_class_uri, default_ok=False, pythonform=True)
            class_class_curie = self.namespaces.curie_for(class_class_uri, default_ok=False, pythonform=False)
        else:
            cls_python_uri = None
            class_class_curie = None
        if class_class_curie:
            class_class_curie = f'"{class_class_curie}"'
        class_class_uri = cls_python_uri if cls_python_uri else f'URIRef("{class_class_uri}")'
        class_model_uri = self.namespaces.uri_or_curie_for(
            self.schema.default_prefix or "DEFAULT_", camelcase(cls.name)
        )
        if ":/" in class_model_uri:
            class_model_uri = f'URIRef("{class_model_uri}")'
        else:
            ns, ln = class_model_uri.split(":", 1)
            class_model_uri = f"{ns.upper()}.{ln}"

        vars = [
            f"class_class_uri: ClassVar[URIRef] = {class_class_uri}",
            f"class_class_curie: ClassVar[str] = {class_class_curie}",
            f'class_name: ClassVar[str] = "{cls.name}"',
            f"class_model_uri: ClassVar[URIRef] = {class_model_uri}",
        ]
        return "\n\t" + "\n\t".join(vars) + "\n"

    def gen_type_meta(self, typ: TypeDefinition) -> str:
        type_class_uri = self.namespaces.uri_for(typ.uri)
        if type_class_uri:
            type_python_uri = self.namespaces.curie_for(type_class_uri, default_ok=False, pythonform=True)
            type_class_curie = self.namespaces.curie_for(type_class_uri, default_ok=False, pythonform=False)
        else:
            type_python_uri = None
            type_class_curie = None
        if type_class_curie:
            type_class_curie = f'"{type_class_curie}"'
        type_class_uri = type_python_uri if type_python_uri else f'URIRef("{type_class_uri}")'
        type_model_uri = self.namespaces.uri_or_curie_for(self.schema.default_prefix, camelcase(typ.name))
        if ":/" in type_model_uri:
            type_model_uri = f'URIRef("{type_model_uri}")'
        else:
            ns, ln = type_model_uri.split(":", 1)
            ln_suffix = f".{ln}" if ln.isidentifier() else f'["{ln}"]'
            type_model_uri = f"{ns.upper()}{ln_suffix}"
        type_meta = [
            f"type_class_uri = {type_class_uri}",
            f"type_class_curie = {type_class_curie}",
            f'type_name = "{typ.name}"',
            f"type_model_uri = {type_model_uri}",
        ]
        return "\n\t".join(type_meta)

    def gen_class_variables(self, cls: ClassDefinition) -> str:
        """
        Generate the variable declarations for a dataclass.

        :param cls: class containing variables to be rendered in inheritance hierarchy
        :return: variable declarations for target class and its ancestors
        """
        initializers = []

        is_root = not cls.is_a
        domain_slots = self.domain_slots(cls)

        # Root keys and identifiers go first.  Note that even if a key or identifier is overridden it still
        # appears at the top of the list, as we need to keep the position
        slot_variables = self._slot_iter(
            cls,
            lambda slot: (slot.identifier or slot.key) and not slot.ifabsent,
            first_hit_only=True,
        )
        initializers += [self.gen_class_variable(cls, slot, not is_root) for slot in slot_variables]

        # Required slots
        slot_variables = self._slot_iter(
            cls,
            lambda slot: slot.required and not slot.identifier and not slot.key and not slot.ifabsent,
        )
        initializers += [self.gen_class_variable(cls, slot, not is_root) for slot in slot_variables]

        # Required or key slots with default values
        slot_variables = self._slot_iter(cls, lambda slot: slot.ifabsent and slot.required)
        initializers += [self.gen_class_variable(cls, slot, False) for slot in slot_variables]

        # Followed by everything else

        slot_variables = self._slot_iter(cls, lambda slot: not slot.required and slot in domain_slots)
        initializers += [self.gen_class_variable(cls, slot, False) for slot in slot_variables]

        return "\n\t".join(initializers)

    def gen_class_variable(self, cls: ClassDefinition, slot: SlotDefinition, can_be_positional: bool = False) -> str:
        """
        Generate a class variable declaration for the supplied slot.  Note: the can_be_positional attribute works,
        but it makes tag/value lists unduly complex, as you can't load them with tag=..., value=... -- you HAVE
        to load positionally. We currently ignore this parameter, meaning that we have a tag/value option for
        any LinkML element

        :param cls: Owning class
        :param slot: slot definition
        :param can_be_positional: True means that positional parameters are allowed.
        :return: Initializer string
        """
        slotname = self.slot_name(slot.name)
        slot_range, default_val = self.range_cardinality(slot, cls, can_be_positional)
        ifabsent_text = self.ifabsent_processor.process_slot(slot, cls) if slot.ifabsent is not None else None
        if ifabsent_text is not None:
            default = f"= {ifabsent_text}"
        else:
            default = f"= {default_val}" if default_val else ""
        return f"""{slotname}: {slot_range} {default}"""

    def range_cardinality(
        self,
        slot: SlotDefinition,
        cls: Optional[ClassDefinition],
        positional_allowed: bool,
    ) -> tuple[str, Optional[str]]:
        """
        Return the range type including initializers, etc.
        Generate a class variable declaration for the supplied slot.  Note: the positional_allowed attribute works,
        but it makes tag/value lists unduly complex, as you can't load them with tag=..., value=... -- you HAVE
        to load positionally. We currently ignore this parameter, meaning that we have a tag/value option for
        any LinkML element

        :param slot: slot to generate type for
        :param cls: containing class -- used to render key slots correctly.  If absent, slot is an add-in
        :param positional_allowed: True Means that we are in the positional space and defaults are not supplied
        :return: python property name and initializer (if any)
        """
        positional_allowed = False  # Force everything to be tag values

        range_type, parent_type, _ = self.class_reference_type(slot, cls)
        pkey = self.class_identifier(slot.range)
        # Special case, inlined, identified range
        if pkey and slot.inlined and slot.multivalued:
            base_key = self.gen_class_reference(self.class_identifier_path(slot.range, False))
            num_elements = len(self.schema.classes[slot.range].slots)
            dflt = None if slot.required and positional_allowed else "empty_dict()"
            if num_elements == 1:
                if slot.required:
                    return (
                        f"Union[list[{base_key}], dict[{base_key}, {range_type}]]",
                        dflt,
                    )
                else:
                    return (
                        f"Optional[Union[list[{base_key}], dict[{base_key}, {range_type}]]]",
                        dflt,
                    )
            else:
                if slot.required:
                    return (
                        f"Union[dict[{base_key}, {range_type}], list[{range_type}]]",
                        dflt,
                    )
                else:
                    return (
                        f"Optional[Union[dict[{base_key}, {range_type}], list[{range_type}]]]",
                        dflt,
                    )

        # All other cases
        if slot.multivalued:
            if slot.required:
                return f"Union[{range_type}, list[{range_type}]]", (None if positional_allowed else "None")
            else:
                return (
                    f"Optional[Union[{range_type}, list[{range_type}]]]",
                    "empty_list()",
                )
        elif slot.required:
            return range_type, (None if positional_allowed else "None")
        else:
            return f"Optional[{range_type}]", "None"

    def class_reference_type(self, slot: SlotDefinition, cls: Optional[ClassDefinition]) -> tuple[str, str, str]:
        """
        Return the type of slot referencing a class

        :param slot: slot to be typed
        :param cls: owning class.  Used for generating key references
        :return: Python class reference type, most proximal type, most proximal type name
        """
        rangelist = (
            self.class_identifier_path(cls, False) if slot.key or slot.identifier else self.slot_range_path(slot)
        )
        prox_type = self.slot_range_path(slot)[-1].rsplit(".")[-1]
        prox_type_name = rangelist[-1]

        # Quote forward references - note that enums always gen at the end
        if slot.range in self.schema.enums or (
            cls and slot.inlined and slot.range in self.schema.classes and self.forward_reference(slot.range, cls.name)
        ):
            rangelist[-1] = f'"{rangelist[-1]}"'
        return str(self.gen_class_reference(rangelist)), prox_type, prox_type_name

    @staticmethod
    def gen_class_reference(rangelist: list[str]) -> str:
        """
        Return a basic or a union type depending on the number of elements in range list

        :param rangelist: List of types from distal to proximal
        :return:
        """
        base = rangelist[0].rsplit(".")[-1]
        return f"Union[{base}, {rangelist[-1]}]" if len(rangelist) > 1 else base

    def gen_postinits(self, cls: ClassDefinition) -> str:
        """Generate all the typing and existence checks post initialize"""
        post_inits = []
        if not (cls.mixin or cls.abstract):
            pkeys = self.primary_keys_for(cls)
            for pkey in pkeys:
                slot = self.schema.slots[pkey]
                # TODO: Remove the bypass whenever we get default_range fixed
                if not slot.ifabsent or True:
                    post_inits.append(self.gen_postinit(cls, slot))
        else:
            pkeys = []
        for slot in self.domain_slots(cls):
            if slot.required:
                # TODO: Remove the bypass whenever we get default_range fixed
                if slot.name not in pkeys and (not slot.ifabsent or True):
                    post_inits.append(self.gen_postinit(cls, slot))
        for slot in self.domain_slots(cls):
            if not slot.required:
                # TODO: Remove the bypass whenever we get default_range fixed
                if slot.name not in pkeys and (not slot.ifabsent or True):
                    post_inits.append(self.gen_postinit(cls, slot))
        post_inits_designators = []

        domain_slot_names = [s.name for s in self.domain_slots(cls)]
        for slot in self.schemaview.class_induced_slots(cls.name):
            # This is for all type designators that were defined at a parent class
            # We need to treat them specially: the initialisation should come
            # AFTER the call to super() because we want to override the super behaviour
            if slot.name not in domain_slot_names and slot.designates_type:
                post_inits_designators.append(self.gen_postinit(cls, slot))

        post_inits_post_super_line = "\n\t\t".join(post_inits_designators)
        post_inits_line = "\n\t\t".join([p for p in post_inits if p])
        return (
            (
                f"""
    def __post_init__(self, *_: str, **kwargs: Any):
        {post_inits_line}
        super().__post_init__(**kwargs)
        {post_inits_post_super_line}"""
            )
            if post_inits_line or post_inits_post_super_line
            else ""
        )

    # sort classes such that if C is a child of P then C appears after P in the list
    @staticmethod
    def _sort_classes(clist: list[ClassDefinition]) -> list[ClassDefinition]:
        clist = list(clist)
        slist = []  # sorted
        while len(clist) > 0:
            for i in range(len(clist)):
                candidate = clist[i]
                can_add = False
                if candidate.is_a is None:
                    can_add = True
                else:
                    if candidate.is_a in [p.name for p in slist]:
                        can_add = True
                if can_add:
                    slist = slist + [candidate]
                    del clist[i]
                    break
            if not can_add:
                raise (f"could not find suitable element in {clist} that does not ref {slist}")
        return slist

    def is_key_value_class(self, range_name: DefinitionName) -> bool:
        """
        Return True if range_name references a class with exactly one key and one value

        :param range_name: class definition (name)
        :return: True if meets the special case
        """
        rng = self.schema.classes.get(range_name)
        if rng:
            pkeys = self.primary_keys_for(rng)
            if pkeys:
                return len(rng.slots) - len(pkeys) == 1
        return False

    def _roll_up_type(self, typ_name: str) -> str:
        if typ_name in self.schemaview.all_types():
            t = self.schemaview.get_type(typ_name)
            if t.typeof:
                return self._roll_up_type(t.typeof)
        return typ_name

    def gen_constructor(self, cls: ClassDefinition) -> Optional[str]:
        """
        Generate python constructor for class

        :param cls: class to generate constructor for
        :return: python constructor
        """
        rlines: list[str] = []
        designators = [x for x in self.domain_slots(cls) if x.designates_type]
        if len(designators) > 0:
            descendants = self.schemaview.class_descendants(cls.name)
            if len(descendants) > 1:
                slot = designators[0]
                aliased_slot_name = self.slot_name(slot.name)
                slot_range = self._roll_up_type(slot.range)

                rlines.append("def __new__(cls, *args, **kwargs):")
                td_val_expression = "kwargs[type_designator]"
                if slot_range == "string":
                    lookup_by_props = ["class_name"]
                elif slot_range == "uri":
                    lookup_by_props = ["class_class_uri", "class_model_uri"]
                    td_val_expression = (
                        f"URIRef({td_val_expression}) if isinstance({td_val_expression}, str) else {td_val_expression}"
                    )
                elif slot_range == "uriorcurie":
                    lookup_by_props = ["class_class_curie", "class_class_uri", "class_model_uri"]
                else:
                    raise ValueError(f"Unsupported type designator range: {slot.range}")
                rlines.append(
                    f"""
        type_designator = "{aliased_slot_name}"
        if not type_designator in kwargs:
            return super().__new__(cls,*args,**kwargs)
        else:
            type_designator_value = {td_val_expression}
            target_cls = cls._class_for("{lookup_by_props[0]}", type_designator_value)
"""
                )
                for prop in lookup_by_props[1:]:
                    rlines.append(
                        f"""
            if target_cls is None:
                target_cls = cls._class_for("{prop}", type_designator_value)
"""
                    )
                rlines.append(
                    f"""
            if target_cls is None:
                raise ValueError(f"Wrong type designator value: class {{cls.__name__}} "
                                 f"has no subclass with {lookup_by_props}='{{kwargs[type_designator]}}'")
            return super().__new__(target_cls,*args,**kwargs)
"""
                )

        if rlines and copy(rlines[-1]).strip() != "":
            rlines.append("")
        return ("\n\t" if len(rlines) > 0 else "") + "\n\t".join(rlines)

    def gen_postinit(self, cls: ClassDefinition, slot: SlotDefinition) -> Optional[str]:
        """Generate python post init rules for slot in class"""
        rlines: list[str] = []

        if slot.range in self.schema.classes:
            if self.is_class_unconstrained(self.schema.classes[slot.range]):
                return ""

        if slot.range in self.schema.enums:
            # Open enum
            if not self.schema.enums[slot.range].permissible_values:
                return ""

        aliased_slot_name = self.slot_name(slot.name)  # Mangled name by which the slot is known in python
        _, _, base_type_name = self.class_reference_type(slot, cls)

        # Generate existence check for required slots.  Note that inherited classes have to do post init checks because
        # You can't have required elements after optional elements in the parent class
        if slot.required:
            rlines.append(f"if self._is_empty(self.{aliased_slot_name}):")
            rlines.append(f'\tself.MissingRequiredField("{aliased_slot_name}")')

        # Generate the type co-ercion for the various types.
        # NOTE: if you set this to true, we will cast all types.   This may be what we really want
        if not slot.multivalued:
            if slot.designates_type:
                pass
            elif slot.required:
                rlines.append(f"if not isinstance(self.{aliased_slot_name}, {base_type_name}):")
            else:
                rlines.append(
                    f"if self.{aliased_slot_name} is not None and "
                    f"not isinstance(self.{aliased_slot_name}, {base_type_name}):"
                )
            if slot.designates_type:
                slot_range = self._roll_up_type(slot.range)
                if slot_range == "string":
                    td_value_classvar = "class_name"
                elif slot_range == "uri":
                    td_value_classvar = "class_model_uri"
                elif slot_range == "uriorcurie":
                    td_value_classvar = "class_class_curie"
                else:
                    raise ValueError(f"Unsupported type designator range: {slot_range}")
                rlines.append(f"self.{aliased_slot_name} = str(self.{td_value_classvar})")
            elif (
                # A really weird case -- a class that has no properties
                slot.range in self.schema.classes and not self.schema.classes[slot.range].slots
            ):
                rlines.append(f"\tself.{aliased_slot_name} = {base_type_name}()")
            else:
                if slot.range in self.schema.enums and slot.ifabsent:
                    # `ifabsent` for an enumeration cannot be assigned to
                    # the dataclass field default, because it would be a
                    # mutable. `python_ifabsent_processor.py` can specify
                    # the default as string and here that string gets
                    # converted into an object attribute invocation
                    # TODO: fix according https://github.com/linkml/linkml/pull/2329#discussion_r1797534588
                    rlines.append(f"\tself.{aliased_slot_name} = getattr({slot.range}, self.{aliased_slot_name})")
                elif (
                    (self.class_identifier(slot.range) and not slot.inlined)
                    or slot.range in self.schema.types
                    or slot.range in self.schema.enums
                ):
                    rlines.append(f"\tself.{aliased_slot_name} = {base_type_name}(self.{aliased_slot_name})")
                else:
                    rlines.append(f"\tself.{aliased_slot_name} = {base_type_name}(**as_dict(self.{aliased_slot_name}))")
        elif slot.inlined:
            slot_range_cls = self.schema.classes[slot.range]
            identifier = self.class_identifier(slot_range_cls)
            # If we don't have an identifier, and we are expecting to be inlined first class elements
            # (inlined_as_list is not True), we will use the first required field as the key.
            #  Note that this may not always work, but the workaround is straight forward -- set inlined_as_list to
            #  True
            if not identifier and not slot.inlined_as_list:
                for range_slot_name in slot_range_cls.slots:
                    range_slot = self.schema.slots[range_slot_name]
                    if range_slot.required:
                        identifier = range_slot.name
                        break
                keyed = False
            else:
                # Place for future expansion
                keyed = True
            if identifier:
                if not slot.inlined_as_list:
                    rlines.append(
                        f'self._normalize_inlined_as_dict(slot_name="{aliased_slot_name}", '
                        f"slot_type={base_type_name}, "
                        f'key_name="{self.aliased_slot_name(identifier)}", '
                        f"keyed={keyed})"
                    )
                else:
                    rlines.append(
                        f'self._normalize_inlined_as_list(slot_name="{aliased_slot_name}", '
                        f"slot_type={base_type_name}, "
                        f'key_name="{self.aliased_slot_name(identifier)}", '
                        f"keyed={keyed})"
                    )
            else:
                # Multivalued, inlined and no identifier
                # TODO: JsonObj([...]) will not be treated correctly here.
                sn = f"self.{aliased_slot_name}"
                rlines.append(f"if not isinstance({sn}, list):")
                rlines.append(f"\t{sn} = [{sn}] if {sn} is not None else []")
                rlines.append(
                    f"{sn} = [v if isinstance(v, {base_type_name}) else {base_type_name}(**as_dict(v)) for v in {sn}]"
                )
        else:
            # Multivalued and not inlined
            # TODO: JsonObj([...]) will fail here as well
            sn = f"self.{aliased_slot_name}"
            rlines.append(f"if not isinstance({sn}, list):")
            rlines.append(f"\t{sn} = [{sn}] if {sn} is not None else []")
            rlines.append(f"{sn} = [v if isinstance(v, {base_type_name}) else {base_type_name}(v) for v in {sn}]")
        while rlines and copy(rlines[-1]).strip() == "":
            rlines.pop()
        rlines.append("")
        return "\n\t\t".join(rlines)

    def _slot_iter(
        self,
        cls: ClassDefinition,
        test: Callable[[SlotDefinition], bool],
        first_hit_only: bool = False,
    ) -> Iterator[SlotDefinition]:
        """Return the representation for the set of own slots in cls that pass test

        :param cls: Class containing a set of slots
        :param test: Slot test function
        :param first_hit_only: True means stop on first match.  False means generate all
        :return: Set of slots that match
        """
        for slot in self.all_slots(cls):
            if test(slot):
                yield slot
                if first_hit_only:
                    break

    def primary_keys_for(self, cls: ClassDefinition) -> list[SlotDefinitionName]:
        """Return the primary key for cls.

        Note: At the moment we return at most one entry.  At some point, keys will be expanded to support
              composite keys.

        @param cls: class to get keys for
        @return: List of primary keys or identifiers
        """
        return [
            slot_name
            for slot_name in cls.slots
            if self.schema.slots[slot_name].key or self.schema.slots[slot_name].identifier
        ]

    def key_name_for(self, class_name: ClassDefinitionName) -> Optional[str]:
        for slot_name in self.primary_keys_for(self.schema.classes[class_name]):
            return self.formatted_element_name(class_name, True) + camelcase(slot_name)
        return None

    def range_type_name(self, slot: SlotDefinition) -> str:
        """Generate the type name for the slot"""
        cidpath = self.slot_range_path(slot)
        if len(cidpath) < 2:
            return cidpath[0]
        else:
            return f"Union[{cidpath[0]}, {cidpath[-1]}]"

    def forward_reference(self, slot_range: str, owning_class: str) -> bool:
        """Determine whether slot_range is a forward reference"""
        # logger.info(f"CHECKING: {slot_range} {owning_class}")
        if (slot_range in self.schema.classes and self.schema.classes[slot_range].imported_from) or (
            slot_range in self.schema.enums and self.schema.enums[slot_range].imported_from
        ):
            logger.info(
                f"FALSE: FORWARD: {slot_range} {owning_class} // IMP={self.schema.classes[slot_range].imported_from}"
            )
            return False
        if slot_range in self.schema.enums:
            return True
        clist = [x.name for x in self._sort_classes(self.schema.classes.values())]
        for cname in clist:
            if cname == owning_class:
                logger.info(f"TRUE: OCCURS SAME: {cname} == {slot_range} owning: {owning_class}")
                return True  # Occurs on or after
            elif cname == slot_range:
                logger.info(f"FALSE: OCCURS BEFORE: {cname} == {slot_range} owning: {owning_class}")
                return False  # Occurs before
        return True

    def python_uri_for(self, uriorcurie: Union[str, URIRef]) -> tuple[str, Optional[str]]:
        """Return the python form of uriorcurie
        :param uriorcurie:
        :return: URI and CURIE form
        """
        ns, ln = self.namespaces.prefix_suffix(uriorcurie)
        if ns == "":
            ns = "DEFAULT_"
        if ns is None:
            return '"str(uriorcurie)"', None
        return (
            ns.upper() + (f".{ln}" if ln.isidentifier() else f"['{ln}']"),
            ns.upper() + f".curie('{ln}')",
        )

    def gen_slotdefs(self) -> str:
        if self.gen_slots:
            return "class slots:\n\tpass\n\n" + "\n\n".join(
                [self.gen_slot(slot) for slot in self.schema.slots.values() if not slot.imported_from]
            )
        else:
            return ""

    def gen_slot(self, slot: SlotDefinition) -> str:
        python_slot_name = underscore(slot.name)
        slot_uri, slot_curie = self.python_uri_for(slot.slot_uri)
        slot_model_uri, slot_model_curie = self.python_uri_for(
            self.namespaces.uri_or_curie_for(self.schema.default_prefix, python_slot_name)
        )
        domain = camelcase(slot.domain) if slot.domain and not self.schema.classes[slot.domain].mixin else "None"
        # Going to omit the range on keys where the domain isn't specified (for now)
        if slot.domain is None and (slot.key or slot.identifier):
            rnge = "URIRef"
        else:
            rnge, _ = self.range_cardinality(slot, self.schema.classes[slot.domain] if slot.domain else None, True)
        if slot.mappings:
            map_texts = [
                self.namespaces.curie_for(self.namespaces.uri_for(m), default_ok=True, pythonform=True)
                for m in slot.mappings
                if m != slot.slot_uri
            ]
        else:
            map_texts = []
        if map_texts:
            mappings = ", mappings = [" + ", ".join(map_texts) + "]"
        else:
            mappings = ""
        pattern = f",\n                   pattern=re.compile(r'{slot.pattern}')" if slot.pattern else ""
        return f"""slots.{python_slot_name} = Slot(uri={slot_uri}, name="{slot.name}", curie={slot_curie},
                   model_uri={slot_model_uri}, domain={domain}, range={rnge}{mappings}{pattern})"""

    def gen_enumerations(self) -> str:
        return "\n\n".join([self.gen_enum(enum) for enum in self.schema.enums.values() if not enum.imported_from])

    def gen_enum(self, enum: EnumDefinition) -> str:
        """
        Generate an enum class
        @param enum: EnumDefinition object to be converted into code
        @return: python code string
        """
        enum_name = camelcase(enum.name)
        return f"""
class {enum_name}(EnumDefinitionImpl):
    {self.gen_enum_comment(enum)}
    {self.gen_enum_description(enum, enum_name)}
""".strip()

    @staticmethod
    def gen_enum_comment(enum: EnumDefinition) -> str:
        if not be(enum.description):
            return ""
        desc_text = enum.description.replace('"""', "---")
        return f'"""\n\t{wrapped_annotation(be(desc_text))}\n\t"""'

    def gen_enum_description(self, enum: EnumDefinition, enum_name: str) -> str:
        return f"""
    {self.gen_pvs(enum)}

    {self.gen_enum_definition(enum, enum_name)}
    {self.gen_pvs_as_setattrs(enum)}
""".strip()

    def gen_enum_definition(self, enum: EnumDefinition, enum_name: str) -> str:
        enum_desc = self.process_multiline_string(enum.description, "\t\tdescription=") if enum.description else None
        desc = f"{enum_desc},\n" if enum.description else ""
        enum_code_set = (
            self.namespaces.curie_for(self.namespaces.uri_for(enum.code_set), default_ok=False, pythonform=True)
            if enum.code_set
            else None
        )
        cs = f"\t\tcode_set={enum_code_set},\n" if enum_code_set else ""
        tag = f'\t\tcode_set_tag="{enum.code_set_tag}",\n' if enum.code_set_tag else ""
        ver = f'\t\tcode_set_version="{enum.code_set_version}",\n' if enum.code_set_version else ""
        vf = f"\t\tpv_formula=PvFormulaOptions.{enum.pv_formula.code.text},\n" if enum.pv_formula else ""

        return f"""_defn = EnumDefinition(\n\t\tname="{enum_name}",\n{desc}{cs}{tag}{ver}{vf}\t)"""

    def gen_pvs(self, enum: EnumDefinition) -> str:
        """
        Generate the python compliant permissible value initializers as a set of class variables
        @param enum: EnumDefinition object to be converted into class variables
        @return: string containing the enum declaration
        """
        init_list = []
        for pv in enum.permissible_values.values():
            if str.isidentifier(pv.text) and not keyword.iskeyword(pv.text):
                init_list.append(f"{pv.text} = " + self.gen_pv_constructor(pv, 4))

        return "\n\t".join(init_list).strip()

    def gen_pvs_as_setattrs(self, enum: EnumDefinition) -> str:
        """
        Generate the non-python compliant permissible value initializers as a set of setattr instructions
        in the form

        @classmethod
        def _addvals(cls):
            setattr(cls, "NAME",
                PermissibleValue(
                    text="NAME",
                    description="description here"))

        @param enum: EnumDefinition object to be converted into code
        @return: string containing the enum declaration
        """
        if any(not str.isidentifier(pv.text) or keyword.iskeyword(pv.text) for pv in enum.permissible_values.values()):
            init_list = []
            for pv in enum.permissible_values.values():
                if not str.isidentifier(pv.text) or keyword.iskeyword(pv.text):
                    # first line is "        setattr("
                    indent = 12
                    indent_str = indent * " "
                    pv_text = pv.text.replace('"', '\\"').replace(r"\n", r"\\n")
                    pv_parts = self.gen_pv_constructor(pv, indent)
                    init_list.append(f'        setattr(cls, "{pv_text}",\n{indent_str}{pv_parts})')

            add_vals_text = "\n".join(init_list).rstrip()

            return f"""
    @classmethod
    def _addvals(cls):
{add_vals_text}
"""

        return ""

    def gen_pv_constructor(self, pv: PermissibleValue, indent: int) -> str:
        """
        Generate a permissible value constructor in the form

        PermissibleValue(text="NAME_ONLY")
        PermissibleValue(
            text="CODE",
            title="...",
            description="...",
            meaning="...")

        @param pv: Value to be constructed
        @param indent: number of additional spaces to add on successive lines
        @return: Permissible value constructor
        """
        constructor = "PermissibleValue"
        pv_text = pv.text.replace('"', '\\"')

        if not pv.description and not pv.meaning and not pv.title:
            return f'{constructor}(text="{pv_text}")'

        indent_str = (4 + indent) * " "
        pv_attrs = [f'{indent_str}text="{pv_text}"']
        if pv.title:
            pv_attrs.append(f'{indent_str}title="{pv.title}"')
        if pv.description:
            pv_attrs.append(f"{self.process_multiline_string(pv.description, f'{indent_str}description=')}")
        if pv.meaning:
            pv_meaning = self.namespaces.curie_for(
                self.namespaces.uri_for(pv.meaning), default_ok=False, pythonform=True
            )
            pv_attrs.append(f"{indent_str}meaning={pv_meaning}")

        return "PermissibleValue(\n" + ",\n".join(pv_attrs) + ")"

    @staticmethod
    def process_multiline_string(input: str, prefix_string: str) -> str:
        """
        Process a (potentially multi-line) string, preserving existing formatting

        @param input: input string to be formatted
        @param prefix_string: the text to prefix the first line of the output
        @return: formatted string
        """
        string = input.rstrip().replace('"', '\\"')
        if len(prefix_string + string) < split_col and input.find("\n") == -1:
            return f'{prefix_string}"{string}"'

        return f'{prefix_string}"""{string}"""'

    def _default_curie_or_uri(self) -> str:
        dflt = self.schema.default_prefix if self.schema.default_prefix else sfx(self.schema.id)
        if ":/" in dflt:
            prefix = self.namespaces.prefix_for(self.schema.default_prefix)
            if prefix:
                dflt = prefix
        return dflt


@shared_arguments(PythonGenerator)
@click.command(name="python")
@click.option("--head/--no-head", default=True, show_default=True, help="Emit metadata heading")
@click.option(
    "--genmeta/--no-genmeta",
    default=False,
    show_default=True,
    help="Generating metamodel. Only use this for generating meta.py",
)
@click.option(
    "--classvars/--no-classvars",
    default=True,
    show_default=True,
    help="Generate CLASSVAR info",
)
@click.option(
    "--slots/--no-slots",
    default=True,
    show_default=True,
    help="Generate Slot information",
)
@click.option(
    "--validate/--no-validate",
    default=False,
    show_default=True,
    help="Validate generated code by compiling it",
)
@click.version_option(__version__, "-V", "--version")
def cli(
    yamlfile,
    head=True,
    genmeta=False,
    classvars=True,
    slots=True,
    validate=False,
    **args,
):
    """Generate python classes to represent a LinkML model"""
    gen = PythonGenerator(
        yamlfile,
        emit_metadata=head,
        genmeta=genmeta,
        gen_classvars=classvars,
        gen_slots=slots,
        **args,
    )
    if validate:
        mod = gen.compile_module()
        logger.info(f"Module {mod} compiled successfully")
    print(gen.serialize(emit_metadata=head, **args))


if __name__ == "__main__":
    cli()



import json
from typing import IO, Any, Optional

from ..config.datamodel.config import RuleLevel
from ..linter import LinterProblem
from .formatter import Formatter


class Encoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, LinterProblem):
            return o.__dict__
        if isinstance(o, RuleLevel):
            return str(o)
        return json.JSONEncoder.default(self, o)


class JsonFormatter(Formatter):
    def __init__(self, file: Optional[IO[Any]] = None) -> None:
        super().__init__(file)
        self.problems = []

    def handle_problem(self, problem: LinterProblem):
        self.problems.append(problem)

    def end_report(self):
        self.write(json.dumps(self.problems, indent=2, cls=Encoder))
